import pytest
from sqlobject import ForeignKey, IntCol, MultipleJoin, SQLMultipleJoin, \
    SQLObject, StringCol
from sqlobject.tests.dbtest import setupClass, supports


class Race(SQLObject):
    name = StringCol()
    fightersAsList = MultipleJoin('RFighter', joinColumn="rf_id")
    fightersAsSResult = SQLMultipleJoin('RFighter', joinColumn="rf_id")


class RFighter(SQLObject):
    name = StringCol()
    race = ForeignKey('Race', dbName="rf_id")
    power = IntCol()


def createAllTables():
    setupClass([Race, RFighter])


def test_1():
    createAllTables()
    # create some races
    human = Race(name='human')
    saiyajin = Race(name='saiyajin')
    hibrid = Race(name='hibrid (human with sayajin)')
    namek = Race(name='namekuseijin')
    # create some fighters
    RFighter(name='Gokou (Kakaruto)', race=saiyajin, power=10)
    RFighter(name='Vegeta', race=saiyajin, power=9)
    RFighter(name='Krilim', race=human, power=3)
    RFighter(name='Yancha', race=human, power=2)
    RFighter(name='Jackie Chan', race=human, power=2)
    RFighter(name='Gohan', race=hibrid, power=8)
    RFighter(name='Goten', race=hibrid, power=7)
    trunks = RFighter(name='Trunks', race=hibrid, power=8)
    picollo = RFighter(name='Picollo', race=namek, power=6)
    RFighter(name='Neil', race=namek, power=5)

    # testing the SQLMultipleJoin stuff
    for i, j in zip(human.fightersAsList, human.fightersAsSResult):
        assert i is j  # the 2 ways should give the same result
    assert namek.fightersAsSResult.count() == len(namek.fightersAsList)
    assert saiyajin.fightersAsSResult.max('power') == 10
    assert trunks in hibrid.fightersAsSResult
    assert picollo not in hibrid.fightersAsSResult
    assert str(hibrid.fightersAsSResult.sum('power')) == '23'


def test_multiple_join_transaction():
    if not supports('transactions'):
        pytest.skip("Transactions aren't supported")
    createAllTables()
    trans = Race._connection.transaction()
    try:
        namek = Race(name='namekuseijin', connection=trans)
        RFighter(name='Gokou (Kakaruto)', race=namek, power=10,
                 connection=trans)
        assert namek.fightersAsSResult.count() == 1
        assert namek.fightersAsSResult[0]._connection == trans
    finally:
        trans.commit(True)
        Race._connection.autoCommit = True



from pytest import raises, skip
from sqlobject import ForeignKey, MultipleJoin, StringCol
from sqlobject.inheritance import InheritableSQLObject
from sqlobject.tests.dbtest import getConnection, setupClass, supports

########################################
# Deep Inheritance
########################################


class DIPerson(InheritableSQLObject):
    firstName = StringCol(length=100)
    lastName = StringCol(alternateID=True, length=100)
    manager = ForeignKey("DIManager", default=None)


class DIEmployee(DIPerson):
    so_position = StringCol(unique=True, length=100)


class DIManager(DIEmployee):
    subdudes = MultipleJoin("DIPerson", joinColumn="manager_id")


def test_creation_fail():
    """
    Try to create a Manager without specifying a position.
    This should fail without leaving any partial records in
    the database.

    """
    setupClass([DIManager, DIEmployee, DIPerson])

    kwargs = {'firstName': 'John', 'lastname': 'Doe'}
    raises(TypeError, DIManager, **kwargs)
    persons = DIEmployee.select(DIPerson.q.firstName == 'John')
    assert persons.count() == 0


def test_creation_fail2():
    """
    Try to create two Managers with the same position.
    This should fail without leaving any partial records in
    the database.

    """
    setupClass([DIManager, DIEmployee, DIPerson])

    kwargs = {'firstName': 'John', 'lastName': 'Doe',
              'so_position': 'Project Manager'}
    DIManager(**kwargs)
    persons = DIEmployee.select(DIPerson.q.firstName == 'John')
    assert persons.count() == 1

    kwargs = {'firstName': 'John', 'lastName': 'Doe II',
              'so_position': 'Project Manager'}
    raises(Exception, DIManager, **kwargs)
    persons = DIPerson.select(DIPerson.q.firstName == 'John')
    assert persons.count() == 1

    if not supports('transactions'):
        skip("Transactions aren't supported")
    transaction = DIPerson._connection.transaction()
    kwargs = {'firstName': 'John', 'lastName': 'Doe III',
              'so_position': 'Project Manager'}
    raises(Exception, DIManager, connection=transaction, **kwargs)
    transaction.rollback()
    transaction.begin()
    persons = DIPerson.select(DIPerson.q.firstName == 'John',
                              connection=transaction)
    assert persons.count() == 1


def test_deep_inheritance():
    conn = getConnection()
    if conn.module.__name__ == 'mysql.connector' \
            and conn.connector_type == 'mysql.connector-python':
        skip("connector-python falls into an infinite loop here")

    setupClass([DIManager, DIEmployee, DIPerson])

    manager = DIManager(firstName='Project', lastName='Manager',
                        so_position='Project Manager')
    manager_id = manager.id
    employee_id = DIEmployee(firstName='Project', lastName='Leader',
                             so_position='Project leader', manager=manager).id
    DIPerson(firstName='Oneof', lastName='Authors', manager=manager)

    cache = conn.cache
    cache.clear()

    managers = list(DIManager.select())
    assert len(managers) == 1
    cache.clear()

    employees = list(DIEmployee.select())
    assert len(employees) == 2
    cache.clear()

    persons = list(DIPerson.select())
    assert len(persons) == 3
    cache.clear()

    person = DIPerson.get(employee_id)
    assert isinstance(person, DIEmployee)

    person = DIPerson.get(manager_id)
    assert isinstance(person, DIEmployee)
    assert isinstance(person, DIManager)
    cache.clear()

    person = DIEmployee.get(manager_id)
    assert isinstance(person, DIManager)
    conn.close()



import pytest
from sqlobject import IntCol, SQLObject
from sqlobject.tests.dbtest import setupClass, supports


########################################
# Slicing tests
########################################


def listrange(*args):
    """Always return a list, for py3k compatibility"""
    return list(range(*args))


class Counter(SQLObject):

    number = IntCol(notNull=True)


class TestSlice:

    def setup_method(self, meth):
        setupClass(Counter)
        for i in range(100):
            Counter(number=i)

    def counterEqual(self, counters, value):
        if not supports('limitSelect'):
            pytest.skip("limitSelect isn't supported")
        assert [c.number for c in counters] == value

    def test_slice(self):
        self.counterEqual(
            Counter.select(None, orderBy='number'), listrange(100))

        self.counterEqual(
            Counter.select(None, orderBy='number')[10:20],
            listrange(10, 20))

        self.counterEqual(
            Counter.select(None, orderBy='number')[20:30][:5],
            listrange(20, 25))

        self.counterEqual(
            Counter.select(None, orderBy='number')[20:30][1:5],
            listrange(21, 25))

        self.counterEqual(
            Counter.select(None, orderBy='number')[:-10],
            listrange(0, 90))

        self.counterEqual(
            Counter.select(None, orderBy='number', reversed=True),
            listrange(99, -1, -1))

        self.counterEqual(
            Counter.select(None, orderBy='-number'),
            listrange(99, -1, -1))



from array import array
import datetime
from decimal import Decimal
import time
import sys
from .compat import PY2, buffer_type

if PY2:
    from types import ClassType, InstanceType, NoneType
else:
    # Use suitable aliases for now
    ClassType = type
    NoneType = type(None)
    # This is may not be what we want in all cases, but will do for now
    InstanceType = object


try:
    from mx.DateTime import DateTimeType as mxDateTimeType, \
        DateTimeDeltaType as mxDateTimeDeltaType
except ImportError:
    mxDateTimeType = None
    mxDateTimeDeltaType = None

try:
    import pendulum
except ImportError:
    pendulumDateTimeType = None
else:
    pendulumDateTimeType = pendulum.DateTime

try:
    from DateTime import DateTime as zopeDateTime
except ImportError:
    zopeDateTimeType = None
else:
    zopeDateTimeType = type(zopeDateTime())

try:
    import Sybase
    NumericType = Sybase.NumericType
except ImportError:
    NumericType = None


########################################
# Quoting
########################################

sqlStringReplace = [
    ("'", "''"),
    ('\\', '\\\\'),
    ('\000', '\\0'),
    ('\b', '\\b'),
    ('\n', '\\n'),
    ('\r', '\\r'),
    ('\t', '\\t'),
]


class ConverterRegistry:

    def __init__(self):
        self.basic = {}
        self.klass = {}

    def registerConverter(self, typ, func):
        if type(typ) is ClassType:
            self.klass[typ] = func
        else:
            self.basic[typ] = func

    if PY2:
        def lookupConverter(self, value, default=None):
            if type(value) is InstanceType:
                # lookup on klasses dict
                return self.klass.get(value.__class__, default)
            return self.basic.get(type(value), default)
    else:
        def lookupConverter(self, value, default=None):
            # python 3 doesn't have classic classes, so everything's
            # in self.klass due to comparison order in registerConvertor
            return self.klass.get(value.__class__, default)

converters = ConverterRegistry()
registerConverter = converters.registerConverter
lookupConverter = converters.lookupConverter


def StringLikeConverter(value, db):
    if isinstance(value, array):
        try:
            value = value.tounicode()
        except ValueError:
            value = value.tostring()
    elif isinstance(value, buffer_type):
        value = str(value)

    if db in ('mysql', 'postgres'):
        for orig, repl in sqlStringReplace:
            value = value.replace(orig, repl)
    elif db in ('sqlite', 'firebird', 'sybase', 'maxdb', 'mssql'):
        value = value.replace("'", "''")
    else:
        assert 0, "Database %s unknown" % db
    if (db == 'postgres') and ('\\' in value):
        return "E'%s'" % value
    return "'%s'" % value

registerConverter(str, StringLikeConverter)
if PY2:
    # noqa for flake8 & python3
    registerConverter(unicode, StringLikeConverter)  # noqa
registerConverter(array, StringLikeConverter)
if PY2:
    registerConverter(bytearray, StringLikeConverter)
    registerConverter(buffer_type, StringLikeConverter)
else:
    registerConverter(memoryview, StringLikeConverter)


def IntConverter(value, db):
    return repr(int(value))

registerConverter(int, IntConverter)


def LongConverter(value, db):
    return str(value)

if sys.version_info[0] < 3:
    # noqa for flake8 & python3
    registerConverter(long, LongConverter)  # noqa

if NumericType:
    registerConverter(NumericType, IntConverter)


def BoolConverter(value, db):
    if db == 'postgres':
        if value:
            return "'t'"
        else:
            return "'f'"
    else:
        if value:
            return '1'
        else:
            return '0'

registerConverter(bool, BoolConverter)


def FloatConverter(value, db):
    return repr(value)

registerConverter(float, FloatConverter)


def NoneConverter(value, db):
    return "NULL"

registerConverter(NoneType, NoneConverter)


def SequenceConverter(value, db):
    return "(%s)" % ", ".join([sqlrepr(v, db) for v in value])

registerConverter(tuple, SequenceConverter)
registerConverter(list, SequenceConverter)
registerConverter(dict, SequenceConverter)
registerConverter(set, SequenceConverter)
registerConverter(frozenset, SequenceConverter)

if hasattr(time, 'struct_time'):
    def StructTimeConverter(value, db):
        return time.strftime("'%Y-%m-%d %H:%M:%S'", value)

    registerConverter(time.struct_time, StructTimeConverter)


def DateTimeConverter(value, db):
    return "'%04d-%02d-%02d %02d:%02d:%02d'" % (
        value.year, value.month, value.day,
        value.hour, value.minute, value.second)


def DateTimeConverterMS(value, db):
    return "'%04d-%02d-%02d %02d:%02d:%02d.%06d'" % (
        value.year, value.month, value.day,
        value.hour, value.minute, value.second, value.microsecond)

registerConverter(datetime.datetime, DateTimeConverterMS)


def DateConverter(value, db):
    return "'%04d-%02d-%02d'" % (value.year, value.month, value.day)

registerConverter(datetime.date, DateConverter)


def TimeConverter(value, db):
    return "'%02d:%02d:%02d'" % (value.hour, value.minute, value.second)


def TimeConverterMS(value, db):
    return "'%02d:%02d:%02d.%06d'" % (value.hour, value.minute,
                                      value.second, value.microsecond)

registerConverter(datetime.time, TimeConverterMS)


if mxDateTimeType:
    def mxDateTimeConverter(value, db):
        return "'%s'" % value.strftime("%Y-%m-%d %H:%M:%S")

    registerConverter(mxDateTimeType, mxDateTimeConverter)

    def mxTimeConverter(value, db):
        return "'%s'" % value.strftime("%H:%M:%S")

    registerConverter(mxDateTimeDeltaType, mxTimeConverter)


if pendulumDateTimeType:
    def pendulumConverter(value, db):
        return "'%s'" % value.to_datetime_string()

    registerConverter(pendulum.DateTime, pendulumConverter)


if zopeDateTimeType:
    def zopeDateTimeConverter(value, db):
        return "'%s'" % value.strftime("%Y-%m-%d %H:%M:%S")

    registerConverter(zopeDateTimeType, zopeDateTimeConverter)


def DecimalConverter(value, db):
    return value.to_eng_string()

registerConverter(Decimal, DecimalConverter)


def TimedeltaConverter(value, db):

    return """INTERVAL '%d days %d seconds'""" % \
        (value.days, value.seconds)

registerConverter(datetime.timedelta, TimedeltaConverter)


def sqlrepr(obj, db=None):
    try:
        reprFunc = obj.__sqlrepr__
    except AttributeError:
        converter = lookupConverter(obj)
        if converter is None:
            raise ValueError("Unknown SQL builtin type: %s for %s" %
                             (type(obj), repr(obj)))
        return converter(obj, db)
    else:
        return reprFunc(db)


def quote_str(s, db):
    if (db == 'postgres') and ('\\' in s):
        return "E'%s'" % s
    return "'%s'" % s


def unquote_str(s):
    if s[:2].upper().startswith("E'") and s.endswith("'"):
        return s[2:-1]
    elif s.startswith("'") and s.endswith("'"):
        return s[1:-1]
    else:
        return s



from sqlobject import BoolCol, FloatCol, IntCol, PickleCol, SQLObject, \
    StringCol, UnicodeCol
from sqlobject.col import validators
from sqlobject.compat import PY2
from sqlobject.tests.dbtest import raises, setupClass

if not PY2:
    # alias for python 3 compatability
    long = int

########################################
# Validation/conversion
########################################


class SOTestValidator(validators.Validator):
    def to_python(self, value, state):
        if value:
            self.save_value.append(value)
            return 1
        return value

    def from_python(self, value, state):
        if value:
            self.save_value.append(value)
            return 2
        return value

validator1 = SOTestValidator(save_value=[])
validator2 = SOTestValidator(save_value=[])


class SOValidation(SQLObject):

    name = StringCol(validator=validators.PlainText(),
                     default='x', dbName='name_col')
    name2 = StringCol(validator2=validators.ConfirmType(type=str), default='y')
    name3 = IntCol(validator=validators.Wrapper(fromPython=int), default=100)
    name4 = FloatCol(default=2.718)
    name5 = PickleCol(default=None)
    name6 = BoolCol(default=None)
    name7 = UnicodeCol(default=None)
    name8 = IntCol(default=None)
    name9 = IntCol(validator=validator1, validator2=validator2, default=0)


class SOValidationTest(object):
    def __init__(self, value):
        self.value = value


if PY2:
    class SOValidationTestUnicode(SOValidationTest):
        def __unicode__(self):
            return self.value


class SOValidationTestInt(SOValidationTest):
    def __int__(self):
        return self.value


class SOValidationTestBool(SOValidationTest):
    def __nonzero__(self):
        return self.value
    __bool__ = __nonzero__


class SOValidationTestFloat(SOValidationTest):
    def __float__(self):
        return self.value


class TestValidation:

    def setup_method(self, meth):
        setupClass(SOValidation)

    def test_validate(self):
        t = SOValidation(name='hey')
        raises(validators.Invalid, setattr, t, 'name', '!!!')
        t.name = 'you'
        assert t.name == 'you'

    def test_confirmType(self):
        t = SOValidation(name2='hey')
        raises(validators.Invalid, setattr, t, 'name2', 1)
        raises(validators.Invalid, setattr, t, 'name3', '1')
        raises(validators.Invalid, setattr, t, 'name4', '1')
        if t._connection.dbName != 'postgres' or \
                t._connection.driver not in ('odbc', 'pyodbc', 'pypyodbc'):
            raises(validators.Invalid, setattr, t, 'name6', '1')
        raises(validators.Invalid, setattr, t, 'name7', 1)
        t.name2 = 'you'
        assert t.name2 == 'you'

        for name, cls, value in (
                ('name4', SOValidationTestFloat, 1.1),
                ('name6', SOValidationTestBool, True),
                ('name8', SOValidationTestInt, 1)):
            setattr(t, name, cls(value))
            assert getattr(t, name) == value
        if PY2:
            for name, cls, value in (
                    ('name7', SOValidationTestUnicode, u'test'),):
                setattr(t, name, cls(value))
                assert getattr(t, name) == value

    def test_wrapType(self):
        t = SOValidation(name3=1)
        raises(validators.Invalid, setattr, t, 'name3', 'x')
        t.name3 = long(1)
        assert t.name3 == 1
        t.name3 = 0
        assert t.name3 == 0

    def test_emptyValue(self):
        t = SOValidation(name5={})
        assert t.name5 == {}

    def test_validator2(self):
        SOValidation(name9=1)
        SOValidation(name9=2)
        assert validator1.save_value == [2, 2, 2, 2, 2, 2]
        assert validator2.save_value == [1, 1, 1, 2, 1, 1]



from sqlobject import sqlbuilder
from sqlobject.classregistry import findClass
from sqlobject.dbconnection import Iteration


class InheritableIteration(Iteration):
    # Default array size for cursor.fetchmany()
    defaultArraySize = 10000

    def __init__(self, dbconn, rawconn, select, keepConnection=False):
        super(InheritableIteration, self).__init__(dbconn, rawconn, select,
                                                   keepConnection)
        self.lazyColumns = select.ops.get('lazyColumns', False)
        try:
            self.cursor.arraysize = self.defaultArraySize
            self.use_arraysize = True
        except AttributeError:  # pymssql doesn't have arraysize
            self.use_arraysize = False
        self._results = []
        # Find the index of the childName column
        childNameIdx = None
        columns = select.sourceClass.sqlmeta.columnList
        for i, column in enumerate(columns):
            if column.name == "childName":
                childNameIdx = i
                break
        self._childNameIdx = childNameIdx

    def next(self):
        if not self._results:
            if self.use_arraysize:
                _results = self.cursor.fetchmany()
            else:
                _results = self.cursor.fetchmany(size=self.defaultArraySize)
            self._results = list(_results)
            if not self.lazyColumns:
                self.fetchChildren()
        if not self._results:
            self._cleanup()
            raise StopIteration
        result = self._results[0]
        del self._results[0]
        if self.lazyColumns:
            obj = self.select.sourceClass.get(result[0],
                                              connection=self.dbconn)
            return obj
        else:
            id = result[0]
            if id in self._childrenResults:
                childResults = self._childrenResults[id]
                del self._childrenResults[id]
            else:
                childResults = None
            obj = self.select.sourceClass.get(
                id, selectResults=result[1:],
                childResults=childResults, connection=self.dbconn)
            return obj

    def fetchChildren(self):
        """Prefetch childrens' data

        Fetch childrens' data for every subclass in one big .select()
        to avoid .get() fetching it one by one.
        """
        self._childrenResults = {}
        if self._childNameIdx is None:
            return
        childIdsNames = {}
        childNameIdx = self._childNameIdx
        for result in self._results:
            childName = result[childNameIdx + 1]
            if childName:
                ids = childIdsNames.get(childName)
                if ids is None:
                    ids = childIdsNames[childName] = []
                ids.append(result[0])
        dbconn = self.dbconn
        rawconn = self.rawconn
        cursor = rawconn.cursor()
        registry = self.select.sourceClass.sqlmeta.registry
        for childName, ids in childIdsNames.items():
            klass = findClass(childName, registry)
            if len(ids) == 1:
                select = klass.select(klass.q.id == ids[0],
                                      childUpdate=True, connection=dbconn)
            else:
                select = klass.select(sqlbuilder.IN(klass.q.id, ids),
                                      childUpdate=True, connection=dbconn)
            query = dbconn.queryForSelect(select)
            if dbconn.debug:
                dbconn.printDebug(rawconn, query,
                                  'Select children of the class %s' %
                                  childName)
            self.dbconn._executeRetry(rawconn, cursor, query)
            for result in cursor.fetchall():
                # Inheritance child classes may have no own columns
                # (that makes sense when child class has a join
                # that does not apply to parent class objects).
                # In such cases result[1:] gives an empty tuple
                # which is interpreted as "no results fetched" in .get().
                # So .get() issues another query which is absolutely
                # meaningless (like "SELECT NULL FROM child WHERE id=1").
                # In order to avoid this, we replace empty results
                # with non-empty tuple.  Extra values in selectResults
                # are Ok - they will be ignored by ._SO_selectInit().
                self._childrenResults[result[0]] = result[1:] or (None,)



import os
from sqlobject.dbconnection import DBConnection
from sqlobject.sqlite.sqliteconnection import SQLiteConnection


########################################
# Test _parseURI
########################################


def test_parse():
    _parseURI = DBConnection._parseURI

    user, password, host, port, path, args = _parseURI("mysql://host/database")
    assert user is None
    assert password is None
    assert host == "host"
    assert port is None
    assert path == "/database"
    assert args == {}

    user, password, host, port, path, args = _parseURI(
        "mysql://user:pass%20word@host/database?unix_socket=/var/mysql/socket")
    assert user == "user"
    assert password == "pass word"
    assert host == "host"
    assert port is None
    assert path == "/database"
    assert args == {"unix_socket": "/var/mysql/socket"}

    user, password, host, port, path, args = \
        _parseURI("postgres://user@host/database")
    assert user == "user"
    assert password is None
    assert host == "host"
    assert port is None
    assert path == "/database"
    assert args == {}

    user, password, host, port, path, args = \
        _parseURI("postgres://host:5432/database")
    assert user is None
    assert password is None
    assert host == "host"
    assert port == 5432
    assert path == "/database"
    assert args == {}

    user, password, host, port, path, args = \
        _parseURI("postgres:///full/path/to/socket/database")
    assert user is None
    assert password is None
    assert host is None
    assert port is None
    assert path == "/full/path/to/socket/database"
    assert args == {}

    user, password, host, port, path, args = \
        _parseURI("postgres://us%3Aer:p%40ssword@host/database")
    assert user == "us:er"
    assert password == "p@ssword"
    assert host == "host"
    assert port is None
    assert path == "/database"
    assert args == {}

    user, password, host, port, path, args = \
        _parseURI("sqlite:///full/path/to/database")
    assert user is None
    assert password is None
    assert host is None
    assert port is None
    assert path == "/full/path/to/database"
    assert args == {}

    user, password, host, port, path, args = _parseURI("sqlite:/:memory:")
    assert user is None
    assert password is None
    assert host is None
    assert port is None
    assert path == "/:memory:"
    assert args == {}

    if os.name == 'nt':
        user, password, host, port, path, args = \
            _parseURI("sqlite:/C|/full/path/to/database")
        assert user is None
        assert password is None
        assert host is None
        assert port is None
        assert path == "C:/full/path/to/database"
        assert args == {}

        user, password, host, port, path, args = \
            _parseURI("sqlite:///C:/full/path/to/database")
        assert user is None
        assert password is None
        assert host is None
        assert port is None
        assert path == "C:/full/path/to/database"
        assert args == {}


def test_uri():
    connection = DBConnection()
    connection.close = lambda: None

    connection.dbName, connection.host, connection.port, \
        connection.user, connection.password, connection.db = \
        'mysql', 'host', None, None, None, 'database'
    assert connection.uri() == "mysql://host/database"

    connection.dbName, connection.host, connection.port, \
        connection.user, connection.password, connection.db = \
        'mysql', 'host', None, 'user', 'pass word', 'database'
    assert connection.uri() == "mysql://user:pass%20word@host/database"

    connection.dbName, connection.host, connection.port, \
        connection.user, connection.password, connection.db = \
        'postgres', 'host', None, 'user', None, 'database'
    assert connection.uri() == "postgres://user@host/database"

    connection.dbName, connection.host, connection.port, \
        connection.user, connection.password, connection.db = \
        'postgres', 'host', 5432, None, None, 'database'
    assert connection.uri() == "postgres://host:5432/database"

    connection.dbName, connection.host, connection.port, \
        connection.user, connection.password, connection.db = \
        'postgres', None, None, None, None, '/full/path/to/socket/database'
    assert connection.uri() == "postgres:///full/path/to/socket/database"

    connection.dbName, connection.host, connection.port, \
        connection.user, connection.password, connection.db = \
        'postgres', 'host', None, 'us:er', 'p@ssword', 'database'
    assert connection.uri() == "postgres://us%3Aer:p%40ssword@host/database"

    connection = SQLiteConnection(None)
    connection.filename = '/full/path/to/database'
    assert connection.uri() == "sqlite:///full/path/to/database"

    connection.filename = ':memory:'
    assert connection.uri() == "sqlite:/:memory:"

    if os.name == 'nt':
        connection.filename = 'C:/full/path/to/database'
        assert connection.uri() == "sqlite:///C%3A/full/path/to/database"



"""
The framework for making database tests.
"""
from __future__ import print_function

import logging
import os
import sys
from pytest import raises, skip
import sqlobject
from sqlobject.col import use_microseconds
import sqlobject.conftest as conftest

if sys.platform[:3] == "win":
    def getcwd():
        return os.getcwd().replace('\\', '/')
else:
    getcwd = os.getcwd

"""
supportsMatrix defines what database backends support what features.
Each feature has a name, if you see a key like '+featureName' then
only the databases listed support the feature.  Conversely,
'-featureName' means all databases *except* the ones listed support
the feature.  The databases are given by their SQLObject string name,
separated by spaces.

The function supports(featureName) returns True or False based on this,
and you can use it like::

    def test_featureX():
        if not supports('featureX'):
            pytest.skip("Doesn't support featureX")
"""
supportsMatrix = {
    '-blobData': 'mssql',
    '-decimalColumn': 'mssql',
    '-dropTableCascade': 'sybase mssql mysql',
    '-emptyTable': 'mssql',
    '+exceptions': 'mysql postgres sqlite',
    '-expressionIndex': 'mysql sqlite firebird mssql',
    '-limitSelect': 'mssql',
    '+memorydb': 'sqlite',
    '+rlike': 'mysql postgres sqlite',
    '+schema': 'postgres',
    '-transactions': ' ',
}


def setupClass(soClasses, force=False):
    """
    Makes sure the classes have a corresponding and correct table.
    This won't recreate the table if it already exists.  It will check
    that the table is properly defined (in case you change your table
    definition).

    You can provide a single class or a list of classes; if a list
    then classes will be created in the order you provide, and
    destroyed in the opposite order.  So if class A depends on class
    B, then do setupClass([B, A]) and B won't be destroyed or cleared
    until after A is destroyed or cleared.

    If force is true, then the database will be recreated no matter
    what.
    """
    global hub
    if not isinstance(soClasses, (list, tuple)):
        soClasses = [soClasses]
    connection = getConnection()
    for soClass in soClasses:
        # This would be an alternate way to register connections:
        # try:
        #    hub
        # except NameError:
        #    hub = sqlobject.dbconnection.ConnectionHub()
        # soClass._connection = hub
        # hub.threadConnection = connection
        # hub.processConnection = connection
        soClass._connection = connection
    installOrClear(soClasses, force=force)
    return soClasses


def speedupSQLiteConnection(connection):
    connection.query("PRAGMA synchronous=OFF")
    connection.query("PRAGMA count_changes=OFF")
    connection.query("PRAGMA journal_mode=MEMORY")
    connection.query("PRAGMA temp_store=MEMORY")

installedDBFilename = os.path.join(getcwd(), 'dbs_data.tmp')

installedDBTracker = sqlobject.connectionForURI(
    'sqlite:///' + installedDBFilename)
speedupSQLiteConnection(installedDBTracker)


def getConnection(**kw):
    name = getConnectionURI()
    conn = sqlobject.connectionForURI(name, **kw)
    if conftest.option.show_sql:
        conn.debug = True
    if conftest.option.show_sql_output:
        conn.debugOutput = True
    if (conn.dbName == 'sqlite') and not conn._memory:
        speedupSQLiteConnection(conn)
    return conn


def getConnectionURI():
    name = conftest.option.Database
    if name in conftest.connectionShortcuts:
        name = conftest.connectionShortcuts[name]
    return name

try:
    connection = getConnection()
except Exception as e:
    # At least this module should be importable...
    # The module was imported during documentation building
    if 'sphinx' not in sys.modules:
        print("Could not open database: %s" % e, file=sys.stderr)
else:
    if (connection.dbName == 'firebird'):
        use_microseconds(False)


class InstalledTestDatabase(sqlobject.SQLObject):
    """
    This table is set up in SQLite (always, regardless of --Database) and
    tracks what tables have been set up in the 'real' database.  This
    way we don't keep recreating the tables over and over when there
    are multiple tests that use a table.
    """

    _connection = installedDBTracker
    table_name = sqlobject.StringCol(notNull=True)
    createSQL = sqlobject.StringCol(notNull=True)
    connectionURI = sqlobject.StringCol(notNull=True)

    @classmethod
    def installOrClear(cls, soClasses, force=False):
        cls.setup()
        reversed = list(soClasses)[:]
        reversed.reverse()
        # If anything needs to be dropped, they all must be dropped
        # But if we're forcing it, then we'll always drop
        if force:
            any_drops = True
        else:
            any_drops = False
        for soClass in reversed:
            table = soClass.sqlmeta.table
            if not soClass._connection.tableExists(table):
                continue
            items = list(cls.selectBy(
                table_name=table,
                connectionURI=soClass._connection.uri()))
            if items:
                instance = items[0]
                sql = instance.createSQL
            else:
                sql = None
            newSQL, constraints = soClass.createTableSQL()
            if sql != newSQL:
                if sql is not None:
                    instance.destroySelf()
                any_drops = True
                break
        for soClass in reversed:
            if soClass._connection.tableExists(soClass.sqlmeta.table):
                if any_drops:
                    cls.drop(soClass)
                else:
                    cls.clear(soClass)
        for soClass in soClasses:
            table = soClass.sqlmeta.table
            if not soClass._connection.tableExists(table):
                cls.install(soClass)

    @classmethod
    def install(cls, soClass):
        """
        Creates the given table in its database.
        """
        sql = getattr(soClass, soClass._connection.dbName + 'Create',
                      None)
        all_extra = []
        if sql:
            soClass._connection.query(sql)
        else:
            sql, extra_sql = soClass.createTableSQL()
            soClass.createTable(applyConstraints=False)
            all_extra.extend(extra_sql)
        cls(table_name=soClass.sqlmeta.table,
            createSQL=sql,
            connectionURI=soClass._connection.uri())
        for extra_sql in all_extra:
            soClass._connection.query(extra_sql)

    @classmethod
    def drop(cls, soClass):
        """
        Drops a the given table from its database
        """
        sql = getattr(soClass, soClass._connection.dbName + 'Drop', None)
        if sql:
            soClass._connection.query(sql)
        else:
            soClass.dropTable()

    @classmethod
    def clear(cls, soClass):
        """
        Removes all the rows from a table.
        """
        soClass.clearTable()

    @classmethod
    def setup(cls):
        """
        This sets up *this* table.
        """
        if not cls._connection.tableExists(cls.sqlmeta.table):
            cls.createTable()

installOrClear = InstalledTestDatabase.installOrClear


class Dummy(object):

    """
    Used for creating fake objects; a really poor 'mock object'.
    """

    def __init__(self, **kw):
        for name, value in kw.items():
            setattr(self, name, value)


def inserts(cls, data, schema=None):
    """
    Creates a bunch of rows.

    You can use it like::

        inserts(Person, [{'fname': 'blah', 'lname': 'doe'}, ...])

    Or::

        inserts(Person, [('blah', 'doe')], schema=
                ['fname', 'lname'])

    If you give a single string for the `schema` then it'll split
    that string to get the list of column names.
    """
    if schema:
        if isinstance(schema, str):
            schema = schema.split()
        keywordData = []
        for item in data:
            itemDict = {}
            for name, value in zip(schema, item):
                itemDict[name] = value
            keywordData.append(itemDict)
        data = keywordData
    results = []
    for args in data:
        results.append(cls(**args))
    return results


def supports(feature):
    dbName = connection.dbName
    support = supportsMatrix.get('+' + feature, None)
    notSupport = supportsMatrix.get('-' + feature, None)
    if support is not None and dbName in support.split():
        return True
    elif support:
        return False
    if notSupport is not None and dbName in notSupport.split():
        return False
    elif notSupport:
        return True
    assert notSupport is not None or support is not None, (
        "The supportMatrix does not list this feature: %r"
        % feature)


# To avoid name clashes:
_inserts = inserts


def setSQLiteConnectionFactory(TableClass, factory):
    from sqlobject.sqlite.sqliteconnection import SQLiteConnection
    conn = TableClass._connection
    TableClass._connection = SQLiteConnection(
        filename=conn.filename,
        name=conn.name, debug=conn.debug, debugOutput=conn.debugOutput,
        cache=conn.cache, style=conn.style, autoCommit=conn.autoCommit,
        debugThreading=conn.debugThreading, registry=conn.registry,
        factory=factory
    )
    speedupSQLiteConnection(TableClass._connection)
    installOrClear([TableClass])


def deprecated_module():
    sqlobject.main.warnings_level = None
    sqlobject.main.exception_level = None


def setup_module(mod):
    # modules with '_old' test backward compatible methods, so they
    # don't get warnings or errors.
    mod_name = str(mod.__name__)
    if mod_name.endswith('/py'):
        mod_name = mod_name[:-3]
    if mod_name.endswith('_old'):
        sqlobject.main.warnings_level = None
        sqlobject.main.exception_level = None
    else:
        sqlobject.main.warnings_level = None
        sqlobject.main.exception_level = 0


def teardown_module(mod=None):
    sqlobject.main.warnings_level = None
    sqlobject.main.exception_level = 0


def setupLogging():
    fmt = '[%(asctime)s] %(name)s %(levelname)s: %(message)s'
    formatter = logging.Formatter(fmt)
    hdlr = logging.StreamHandler(sys.stderr)
    hdlr.setFormatter(formatter)
    hdlr.setLevel(logging.NOTSET)
    logger = logging.getLogger()
    logger.addHandler(hdlr)


def setupCyclicClasses(*classes):
    if not supports('dropTableCascade'):
        skip("dropTableCascade isn't supported")
    conn = getConnection()
    for soClass in classes:
        soClass.setConnection(conn)
        soClass.dropTable(ifExists=True, cascade=True)

    constraints = []
    for soClass in classes:
        constraints += soClass.createTable(ifNotExists=True,
                                           applyConstraints=False)
    for constraint in constraints:
        conn.query(constraint)


__all__ = ['Dummy', 'deprecated_module', 'getConnection', 'getConnectionURI',
           'inserts', 'raises', 'setupClass', 'setupCyclicClasses',
           'setupLogging', 'setup_module', 'supports', 'teardown_module',
           ]



import sqlobject.col
from sqlobject.compat import PY2, string_type


__all__ = ['HashCol']


class DbHash:
    """
    Presents a comparison object for hashes, allowing plain text to be
    automagically compared with the base content
    """

    def __init__(self, hash, hashMethod):
        self.hash = hash
        self.hashMethod = hashMethod

    def _get_key(self, other):
        """Create the hash of the other class"""
        if not isinstance(other, string_type):
            raise TypeError(
                "A hash may only be compared with a string, or None.")
        return self.hashMethod(other)

    def __eq__(self, other):
        if other is None:
            if self.hash is None:
                return True
            return False
        other_key = self._get_key(other)
        return other_key == self.hash

    def __lt__(self, other):
        if other is None:
            return False
        other_key = self._get_key(other)
        return other_key < self.hash

    def __gt__(self, other):
        if other is None:
            if self.hash is None:
                return False
            return True
        other_key = self._get_key(other)
        return other_key > self.hash

    def __le__(self, other):
        if other is None:
            if self.hash is None:
                return True
            return False
        other_key = self._get_key(other)
        return other_key <= self.hash

    def __ge__(self, other):
        if other is None:
            return True
        other_key = self._get_key(other)
        return other_key >= self.hash

    def __repr__(self):
        return "<DbHash>"


class HashValidator(sqlobject.col.StringValidator):
    """Provides formal SQLObject validation services for the HashCol"""

    def to_python(self, value, state):
        """ Passes out a hash object. """
        if value is None:
            return None
        return DbHash(hash=value, hashMethod=self.hashMethod)

    def from_python(self, value, state):
        """Store the given value as a MD5 hash, or None if specified"""
        if value is None:
            return None
        return self.hashMethod(value)


class SOHashCol(sqlobject.col.SOStringCol):
    """The internal HashCol definition. By default, enforces a md5 digest"""

    def __init__(self, **kw):
        if 'hashMethod' not in kw:
            from hashlib import md5
            if PY2:
                self.hashMethod = lambda v: md5(v).hexdigest()
            else:
                self.hashMethod = lambda v: md5(v.encode('utf8')).hexdigest()
            if 'length' not in kw:
                kw['length'] = 32
        else:
            self.hashMethod = kw['hashMethod']
            del kw['hashMethod']
        super(SOHashCol, self).__init__(**kw)

    def createValidators(self):
        return [HashValidator(name=self.name, hashMethod=self.hashMethod)] + \
            super(SOHashCol, self).createValidators()


class HashCol(sqlobject.col.StringCol):
    """
    End-user HashCol class. May be instantiated with 'hashMethod', a function
    which returns the string hash of any other string (i.e. basestring)
    """

    baseClass = SOHashCol



# These tests are not enabled yet, but here are some working examples
# of using createSQL so far.

from sqlobject import SQLObject, StringCol


class SOTest1(SQLObject):
    class sqlmeta:
        createSQL = "CREATE SEQUENCE db_test1_seq;"
    test1 = StringCol()


class SOTest2(SQLObject):
    class sqlmeta:
        createSQL = ["CREATE SEQUENCE db_test2_seq;",
                     "ALTER TABLE test2 ADD CHECK(test2 != '');"]
    test2 = StringCol()


class SOTest3(SQLObject):
    class sqlmeta:
        createSQL = {'postgres': 'CREATE SEQUENCE db_test3_seq;',
                     'mysql': 'CREATE SEQUENCE db_test3_seq;'}
    test3 = StringCol()


class SOTest4(SQLObject):
    class sqlmeta:
        createSQL = {'postgres': ['CREATE SEQUENCE db_test4_seq;',
                                  "ALTER TABLE test4 ADD CHECK(test4 != '');"],
                     'mysql': 'CREATE SEQUENCE db_test4_seq;'}
    test4 = StringCol()


class SOTest5(SQLObject):
    class sqlmeta:
        createSQL = {'mysql': 'CREATE SEQUENCE db_test5_seq;'}
    test5 = StringCol()



from sqlobject import SQLObject, StringCol, sqlmeta
from sqlobject.tests.dbtest import setupClass


########################################
# String ID test
########################################


class SOStringID(SQLObject):

    class sqlmeta(sqlmeta):
        table = 'so_string_id'
        idType = str
    val = StringCol(alternateID=True)

    mysqlCreate = """
    CREATE TABLE IF NOT EXISTS so_string_id (
      id VARCHAR(50) PRIMARY KEY,
      val TEXT
    )
    """

    postgresCreate = """
    CREATE TABLE so_string_id (
      id VARCHAR(50) PRIMARY KEY,
      val TEXT
    )
    """

    sybaseCreate = """
    CREATE TABLE so_string_id (
      id VARCHAR(50) UNIQUE,
      val VARCHAR(50) NULL
    )
    """

    firebirdCreate = """
    CREATE TABLE so_string_id (
      id VARCHAR(50) NOT NULL PRIMARY KEY,
      val BLOB SUB_TYPE TEXT
    )
    """

    mssqlCreate = """
    CREATE TABLE so_string_id (
      id VARCHAR(50) PRIMARY KEY,
      val varchar(4000)
    )
    """

    sqliteCreate = postgresCreate

    mysqlDrop = """
    DROP TABLE IF EXISTS so_string_id
    """

    postgresDrop = """
    DROP TABLE so_string_id
    """

    sqliteDrop = postgresDrop
    firebirdDrop = postgresDrop
    mssqlDrop = postgresDrop


def test_stringID():
    setupClass(SOStringID)
    t1 = SOStringID(id='hey', val='whatever')
    t2 = SOStringID.byVal('whatever')
    assert t1 == t2
    assert t1.val == t2.val
    assert t1.val == 'whatever'

    t1 = SOStringID(id='you', val='nowhere')
    t2 = SOStringID.get('you')
    assert t1 == t2
    assert t1.val == t2.val
    assert t1.val == 'nowhere'



from sqlobject import SQLObject, StringCol
from sqlobject.tests.dbtest import setupClass


########################################
# Inheritance
########################################


class Super(SQLObject):

    name = StringCol(length=10)


class Sub(Super):

    name2 = StringCol(length=10)


def test_super():
    setupClass(Super)
    setupClass(Sub)
    s1 = Super(name='one')
    Super(name='two')  # s2
    s3 = Super.get(s1.id)
    assert s1 == s3


def test_sub():
    setupClass(Super)
    setupClass(Sub)
    s1 = Sub(name='one', name2='1')
    Sub(name='two', name2='2')  # s2
    s3 = Sub.get(s1.id)
    assert s1 == s3



import sys
import sqlobject
from sqlobject.compat import string_type

# The module was imported during documentation building
if 'sphinx' not in sys.modules:
    from paste.deploy.converters import asbool
    from paste.wsgilib import catch_errors
    from paste.util import import_string


def make_middleware(app, global_conf, database=None, use_transaction=False,
                    hub=None):
    """
    WSGI middleware that sets the connection for the request (using
    the database URI or connection object) and the given hub (or
    ``sqlobject.sqlhub`` if not given).

    If ``use_transaction`` is true, then the request will be run in a
    transaction.

    Applications can use the keys (which are all no-argument functions):

    ``sqlobject.get_connection()``:
      Returns the connection object

    ``sqlobject.abort()``:
      Aborts the transaction.  Does not raise an error, but at the *end*
      of the request there will be a rollback.

    ``sqlobject.begin()``:
      Starts a transaction.  First commits (or rolls back if aborted) if
      this is run in a transaction.

    ``sqlobject.in_transaction()``:
      Returns true or false, depending if we are currently in a
      transaction.
    """
    use_transaction = asbool(use_transaction)
    if database is None:
        database = global_conf.get('database')
    if not database:
        raise ValueError(
            "You must provide a 'database' configuration value")
    if isinstance(hub, string_type):
        hub = import_string.eval_import(hub)
    if not hub:
        hub = sqlobject.sqlhub
    if isinstance(database, string_type):
        database = sqlobject.connectionForURI(database)
    return SQLObjectMiddleware(app, database, use_transaction, hub)


class SQLObjectMiddleware(object):

    def __init__(self, app, conn, use_transaction, hub):
        self.app = app
        self.conn = conn
        self.use_transaction = use_transaction
        self.hub = hub

    def __call__(self, environ, start_response):
        conn = [self.conn]
        if self.use_transaction:
            conn[0] = conn[0].transaction()
        any_errors = []
        use_transaction = [self.use_transaction]
        self.hub.threadConnection = conn[0]

        def abort():
            assert use_transaction[0], (
                "You cannot abort, because a transaction is not being used")
            any_errors.append(None)

        def begin():
            if use_transaction[0]:
                if any_errors:
                    conn[0].rollback()
                else:
                    conn[0].commit()
            any_errors[:] = []
            use_transaction[0] = True
            conn[0] = self.conn.transaction()
            self.hub.threadConnection = conn[0]

        def error(exc_info=None):
            any_errors.append(None)
            ok()

        def ok():
            if use_transaction[0]:
                if any_errors:
                    conn[0].rollback()
                else:
                    conn[0].commit(close=True)
            self.hub.threadConnection = None

        def in_transaction():
            return use_transaction[0]

        def get_connection():
            return conn[0]

        environ['sqlobject.get_connection'] = get_connection
        environ['sqlobject.abort'] = abort
        environ['sqlobject.begin'] = begin
        environ['sqlobject.in_transaction'] = in_transaction
        return catch_errors(self.app, environ, start_response,
                            error_callback=error, ok_callback=ok)



import os
import re
import warnings
from sqlobject import col
from sqlobject.dbconnection import DBAPI


class FirebirdConnection(DBAPI):

    supportTransactions = False
    dbName = 'firebird'
    schemes = [dbName]

    limit_re = re.compile(r'^\s*(select )(.*)', re.IGNORECASE)

    def __init__(self, host, db, port='3050', user='sysdba',
                 password='masterkey', autoCommit=1,
                 dialect=None, role=None, charset=None, **kw):
        drivers = kw.pop('driver', None) or 'fdb,kinterbasdb'
        for driver in drivers.split(','):
            driver = driver.strip()
            if not driver:
                continue
            try:
                if driver == 'fdb':
                    import fdb
                    self.module = fdb
                elif driver == 'kinterbasdb':
                    import kinterbasdb
                    # See
                    # http://kinterbasdb.sourceforge.net/dist_docs/usage.html
                    # for an explanation; in short: use datetime, decimal and
                    # unicode.
                    kinterbasdb.init(type_conv=200)
                    self.module = kinterbasdb
                elif driver in ('firebirdsql', 'pyfirebirdsql'):
                    import firebirdsql
                    self.module = firebirdsql
                else:
                    raise ValueError(
                        'Unknown FireBird driver "%s", '
                        'expected fdb, kinterbasdb or firebirdsql' % driver)
            except ImportError:
                pass
            else:
                break
        else:
            raise ImportError(
                'Cannot find an FireBird driver, tried %s' % drivers)
        self.host = host
        self.port = port
        self.db = db
        self.user = user
        self.password = password
        if dialect:
            self.dialect = int(dialect)
        else:
            self.dialect = None
        self.role = role
        if charset:
            # Encoding defined by user in the connection string.
            self.dbEncoding = charset.replace('-', '')
        else:
            self.dbEncoding = charset
        # Encoding defined during database creation and stored in the database.
        self.defaultDbEncoding = ''
        DBAPI.__init__(self, **kw)

    @classmethod
    def _connectionFromParams(cls, auth, password, host, port, path, args):
        if not password:
            password = 'masterkey'
        if not auth:
            auth = 'sysdba'
        # check for alias using
        if (path[0] == '/') and path[-3:].lower() not in ('fdb', 'gdb'):
            path = path[1:]
        path = path.replace('/', os.sep)
        return cls(host, port=port, db=path, user=auth, password=password,
                   **args)

    def _runWithConnection(self, meth, *args):
        if not self.autoCommit:
            return DBAPI._runWithConnection(self, meth, args)
        conn = self.getConnection()
        # @@: Horrible auto-commit implementation.  Just horrible!
        try:
            conn.begin()
        except self.module.ProgrammingError:
            pass
        try:
            val = meth(conn, *args)
            try:
                conn.commit()
            except self.module.ProgrammingError:
                pass
        finally:
            self.releaseConnection(conn)
        return val

    def _setAutoCommit(self, conn, auto):
        # Only _runWithConnection does "autocommit", so we don't
        # need to worry about that.
        pass

    def makeConnection(self):
        extra = {}
        if self.dialect:
            extra['dialect'] = self.dialect
        return self.module.connect(
            host=self.host,
            port=self.port,
            database=self.db,
            user=self.user,
            password=self.password,
            role=self.role,
            charset=self.dbEncoding,
            **extra
        )

    def _queryInsertID(self, conn, soInstance, id, names, values):
        """Firebird uses 'generators' to create new ids for a table.
        The users needs to create a generator named GEN_<tablename>
        for each table this method to work."""
        table = soInstance.sqlmeta.table
        idName = soInstance.sqlmeta.idName
        sequenceName = soInstance.sqlmeta.idSequence or 'GEN_%s' % table
        c = conn.cursor()
        if id is None:
            c.execute('SELECT gen_id(%s,1) FROM rdb$database' % sequenceName)
            id = c.fetchone()[0]
        names = [idName] + names
        values = [id] + values
        q = self._insertSQL(table, names, values)
        if self.debug:
            self.printDebug(conn, q, 'QueryIns')
        c.execute(q)
        c.close()
        if self.debugOutput:
            self.printDebug(conn, id, 'QueryIns', 'result')
        return id

    @classmethod
    def _queryAddLimitOffset(cls, query, start, end):
        """Firebird slaps the limit and offset (actually 'first' and
        'skip', respectively) statement right after the select."""
        if not start:
            limit_str = "SELECT FIRST %i" % end
        if not end:
            limit_str = "SELECT SKIP %i" % start
        else:
            limit_str = "SELECT FIRST %i SKIP %i" % (end - start, start)

        match = cls.limit_re.match(query)
        if match and len(match.groups()) == 2:
            return ' '.join([limit_str, match.group(2)])
        else:
            return query

    def createTable(self, soClass):
        self.query('CREATE TABLE %s (\n%s\n)' %
                   (soClass.sqlmeta.table, self.createColumns(soClass)))
        self.query("CREATE GENERATOR GEN_%s" % soClass.sqlmeta.table)
        return []

    def createReferenceConstraint(self, soClass, col):
        return None

    def createColumn(self, soClass, col):
        return col.firebirdCreateSQL()

    def createIDColumn(self, soClass):
        key_type = {int: "INT", str: "VARCHAR(255)"}[soClass.sqlmeta.idType]
        return '%s %s NOT NULL PRIMARY KEY' % (soClass.sqlmeta.idName,
                                               key_type)

    def createIndexSQL(self, soClass, index):
        return index.firebirdCreateIndexSQL(soClass)

    def joinSQLType(self, join):
        return 'INT NOT NULL'

    def tableExists(self, tableName):
        # there's something in the database by this name...let's
        # assume it's a table.  By default, fb 1.0 stores EVERYTHING
        # it cares about in uppercase.
        result = self.queryOne(
            "SELECT COUNT(rdb$relation_name) FROM rdb$relations "
            "WHERE rdb$relation_name = '%s'" % tableName.upper())
        return result[0]

    def addColumn(self, tableName, column):
        self.query('ALTER TABLE %s ADD %s' %
                   (tableName,
                    column.firebirdCreateSQL()))

    def dropTable(self, tableName, cascade=False):
        self.query("DROP TABLE %s" % tableName)
        self.query("DROP GENERATOR GEN_%s" % tableName)

    def delColumn(self, sqlmeta, column):
        self.query('ALTER TABLE %s DROP %s' % (sqlmeta.table, column.dbName))

    def readDefaultEncodingFromDB(self):
        # Get out if encoding is known allready (can by None as well).
        if self.defaultDbEncoding == "":
            self.defaultDbEncoding = str(self.queryOne(
                "SELECT rdb$character_set_name FROM rdb$database")[0].
                strip().lower())  # encoding defined during db creation
            if self.defaultDbEncoding == "none":
                self.defaultDbEncoding = None
            if self.dbEncoding != self.defaultDbEncoding:
                warningText = """\n
                   Database charset: %s is different """ \
                   """from connection charset: %s.\n""" % (
                   self.defaultDbEncoding, self.dbEncoding)
                warnings.warn(warningText)
                # TODO: ??? print out the uri string,
                # so user can see what is going on
                warningText = """\n
                   Every CHAR or VARCHAR field can (or, better: must) """ \
                   """have a character set defined in Firebird.
                   In the case, field charset is not defined, """ \
                   """SQLObject try to use a db default encoding instead.
                   Firebird is unable to transliterate between character sets.
                   So you must set the correct values on the server """ \
                   "and on the client if everything is to work smoothely.\n"
                warnings.warn(warningText)

            if not self.dbEncoding:  # defined by user in the connection string
                self.dbEncoding = self.defaultDbEncoding
                warningText = """\n
                   encoding: %s will be used as default """ \
                   """for this connection\n""" % self.dbEncoding
                warnings.warn(warningText)

    def columnsFromSchema(self, tableName, soClass):
        """
        Look at the given table and create Col instances (or
        subclasses of Col) for the fields it finds in that table.
        """

        self.readDefaultEncodingFromDB()

        fieldQuery = """\
        SELECT r.RDB$FIELD_NAME AS field_name,
                CASE f.RDB$FIELD_TYPE
                when 7 then 'smallint'
                when 8 then 'integer'
                when 16 then 'int64'
                when 9 then 'quad'
                when 10 then 'float'
                when 11 then 'd_float'
                when 17 then 'boolean'
                when 27 then 'double'
                when 12 then 'date'
                when 13 then 'time'
                when 35 then 'timestamp'
                when 261 then 'blob'
                when 37 then 'varchar'
                when 14 then 'char'
                when 40 then 'cstring'
                when 45 then 'blob_id'
                  ELSE 'UNKNOWN'
                END AS field_type,
                case f.rdb$field_type
                when 7 then
                case f.rdb$field_sub_type
                    when 1 then 'numeric'
                    when 2 then 'decimal'
                end
                when 8 then
                case f.rdb$field_sub_type
                    when 1 then 'numeric'
                    when 2 then 'decimal'
                end
                when 16 then
                case f.rdb$field_sub_type
                    when 1 then 'numeric'
                    when 2 then 'decimal'
                    else 'bigint'
                end
                when 14 then
                case f.rdb$field_sub_type
                    when 0 then 'unspecified'
                    when 1 then 'binary'
                    when 3 then 'acl'
                    else
                    case
                        when f.rdb$field_sub_type is null then 'unspecified'
                    end
                end
                when 37 then
                case f.rdb$field_sub_type
                    when 0 then 'unspecified'
                    when 1 then 'text'
                    when 3 then 'acl'
                    else
                    case
                        when f.rdb$field_sub_type is null then 'unspecified'
                    end
                end
                when 261 then
                case f.rdb$field_sub_type
                    when 0 then 'unspecified'
                    when 1 then 'text'
                    when 2 then 'blr'
                    when 3 then 'acl'
                    when 4 then 'reserved'
                    when 5 then 'encoded-meta-data'
                    when 6 then 'irregular-finished-multi-db-tx'
                    when 7 then 'transactional_description'
                    when 8 then 'external_file_description'
                end
            end as "ActualSubType",
                f.RDB$FIELD_LENGTH AS field_length,
                f.RDB$FIELD_PRECISION AS field_precision,
                f.RDB$FIELD_SCALE AS field_scale,
                cset.RDB$CHARACTER_SET_NAME AS field_charset,
                coll.RDB$COLLATION_NAME AS field_collation,
                r.rdb$default_source,
                r.RDB$NULL_FLAG AS field_not_null_constraint,
                r.RDB$DESCRIPTION AS field_description
           FROM RDB$RELATION_FIELDS r
           LEFT JOIN RDB$FIELDS f ON r.RDB$FIELD_SOURCE = f.RDB$FIELD_NAME
           LEFT JOIN RDB$COLLATIONS coll
                ON f.RDB$COLLATION_ID = coll.RDB$COLLATION_ID
           LEFT JOIN RDB$CHARACTER_SETS cset
                ON f.RDB$CHARACTER_SET_ID = cset.RDB$CHARACTER_SET_ID
          WHERE r.RDB$RELATION_NAME='%s'  -- table name
        ORDER BY r.RDB$FIELD_POSITION
        """

        colData = self.queryAll(fieldQuery % tableName.upper())
        results = []
        for (field, fieldType, fieldSubtype, fieldLength, fieldPrecision,
                fieldScale, fieldCharset, collationName, defaultSource,
                fieldNotNullConstraint, fieldDescription) in colData:
            field = field.strip().lower()
            fieldType = fieldType.strip()
            if fieldCharset:
                fieldCharset = str(fieldCharset.strip())
                # 'UNICODE_FSS' is less strict
                # Firebird/Interbase UTF8 definition
                if fieldCharset.startswith('UNICODE_FSS'):
                    fieldCharset = "UTF8"
            if fieldSubtype:
                fieldSubtype = fieldSubtype.strip()
                if fieldType == "int64":
                    fieldType = fieldSubtype

            # can look like: "DEFAULT 0", "DEFAULT 'default text'", None
            if defaultSource:
                defaultSource = defaultSource.split(' ')[1]
                if defaultSource.startswith("'") and \
                        defaultSource.endswith("'"):
                    defaultSource = str(defaultSource[1:-1])
                elif fieldType in ("integer", "smallint", "bigint"):
                    defaultSource = int(defaultSource)
                elif fieldType in ("float", "double"):
                    defaultSource = float(defaultSource)
            # TODO: other types for defaultSource
            #    elif fieldType == "datetime":

            idName = str(soClass.sqlmeta.idName or 'id').upper()
            if field.upper() == idName:
                continue
            if fieldScale:
                # PRECISION refers to the total number of digits,
                # and SCALE refers to the number of digits
                # to the right of the decimal point.
                # Both numbers can be from 1 to 18 (SQL dialect 1: 1-15),
                # but SCALE mustbe less than or equal to PRECISION.
                if fieldScale > fieldLength:
                    fieldScale = fieldLength
            colClass, kw = self.guessClass(fieldType, fieldLength,
                                           fieldCharset, fieldScale)
            kw['name'] = str(
                soClass.sqlmeta.style.dbColumnToPythonAttr(field).strip())
            kw['dbName'] = str(field)
            kw['notNone'] = not fieldNotNullConstraint
            kw['default'] = defaultSource
            results.append(colClass(**kw))
        return results

    def guessClass(self, t, flength, fCharset, fscale=None):
        """
        An internal method that tries to figure out what Col subclass
        is appropriate given whatever introspective information is
        available -- both very database-specific.
        """

        # TODO: check if negative values are allowed for fscale

        if t == 'smallint':        # -32,768 to +32,767, 16 bits
            return col.IntCol, {}
        elif t == 'integer':       # -2,147,483,648 to +2,147,483,647, 32 bits
            return col.IntCol, {}
        elif t == 'bigint':
            # -2^63 to 2^63-1 or
            # -9,223,372,036,854,775,808 to 9,223,372,036,854,775,807, 64 bits
            return col.IntCol, {}
        elif t == 'float':
            # 32 bits, 3.4x10^-38 to 3.4x10^38, 7 digit precision
            # (7 significant decimals)
            return col.FloatCol, {}
        elif t == 'double':
            # 64 bits, 1.7x10^-308 to 1.7x10^308, 15 digit precision
            # (15 significant decimals)
            return col.FloatCol, {}
        elif t == 'numeric':
            # Numeric and Decimal are internally stored as smallint,
            # integer or bigint depending on the size.
            # They can handle up to 18 digits.
            if (not flength or not fscale):
                # If neither PRECISION nor SCALE are specified,
                # Firebird/InterBase defines the column as INTEGER
                # instead of NUMERIC and stores only the integer portion
                # of the value
                return col.IntCol, {}
            # check if negative values are allowed for fscale
            return col.DecimalCol, {'size': flength, 'precision': fscale}

        elif t == 'decimal':
            # Check if negative values are allowed for fscale
            return col.DecimalCol, {'size': flength, 'precision': fscale}
        elif t == 'date':  # 32 bits, 1 Jan 100. to 29 Feb 32768.
            return col.DateCol, {}
        elif t == 'time':  # 32 bits, 00:00 to 23:59.9999
            return col.TimeCol, {}
        elif t == 'timestamp':  # 64 bits, 1 Jan 100 to 28 Feb 32768.
            return col.DateTimeCol, {}
        elif t == 'char':  # 32767 bytes
            if fCharset and (fCharset != "NONE"):
                return col.UnicodeCol, {'length': flength, 'varchar': False,
                                        'dbEncoding': fCharset}
            elif self.dbEncoding:
                return col.UnicodeCol, {'length': flength, 'varchar': False,
                                        'dbEncoding': self.dbEncoding}
            else:
                return col.StringCol, {'length': flength, 'varchar': False}
        elif t == 'varchar':  # 32767 bytes
            if fCharset and (fCharset != "NONE"):
                return col.UnicodeCol, {'length': flength, 'varchar': True,
                                        'dbEncoding': fCharset}
            elif self.dbEncoding:
                return col.UnicodeCol, {'length': flength, 'varchar': True,
                                        'dbEncoding': self.dbEncoding}
            else:
                return col.StringCol, {'length': flength, 'varchar': True}

        elif t == 'blob':  # 32GB
            return col.BLOBCol, {}
        else:
            return col.Col, {}

    def createEmptyDatabase(self):
        self.module.create_database(
            "CREATE DATABASE '%s' user '%s' password '%s'" %
            (self.db, self.user, self.password))

    def dropDatabase(self):
        self.module.drop_database()



"""SQLObject"""

# Do import for namespace
# noqa is a directive for flake8 to ignore seemingly unused imports

from .__version__ import version, version_info  # noqa

from .col import *  # noqa
from .index import *  # noqa
from .joins import *  # noqa
from .main import *  # noqa
from .sqlbuilder import AND, OR, NOT, IN, LIKE, RLIKE, DESC, CONTAINSSTRING, const, func  # noqa
from .styles import *  # noqa
from .dbconnection import connectionForURI  # noqa
from . import dberrors  # noqa



from sqlobject import IntCol
from sqlobject.inheritance import InheritableSQLObject
from sqlobject.tests.dbtest import raises, setupClass


class SOTestAggregate1(InheritableSQLObject):
    value1 = IntCol()


class SOTestAggregate2(SOTestAggregate1):
    value2 = IntCol()


def test_aggregates():
    setupClass([SOTestAggregate1, SOTestAggregate2])

    SOTestAggregate1(value1=1)
    SOTestAggregate2(value1=2, value2=12)

    assert SOTestAggregate1.select().max("value1") == 2
    assert SOTestAggregate2.select().max("value1") == 2
    raises(Exception, SOTestAggregate2.select().max, "value2")
    assert SOTestAggregate2.select().max(SOTestAggregate2.q.value2) == 12



import os
import sys
from sqlobject.compat import load_module_from_file


def load_module(module_name):
    mod = __import__(module_name)
    components = module_name.split('.')
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod


def load_module_from_name(filename, module_name):
    if module_name in sys.modules:
        return sys.modules[module_name]
    init_filename = os.path.join(os.path.dirname(filename), '__init__.py')
    if not os.path.exists(init_filename):
        try:
            f = open(init_filename, 'w')
        except (OSError, IOError) as e:
            raise IOError(
                'Cannot write __init__.py file into directory %s (%s)\n'
                % (os.path.dirname(filename), e))
        f.write('#\n')
        f.close()
    if module_name in sys.modules:
        return sys.modules[module_name]
    if '.' in module_name:
        parent_name = '.'.join(module_name.split('.')[:-1])
        base_name = module_name.split('.')[-1]
        load_module_from_name(os.path.dirname(filename), parent_name)
    else:
        base_name = module_name
    return load_module_from_file(base_name, module_name, filename)



from sqlobject.dbconnection import registerConnection


def builder():
    from . import firebirdconnection
    return firebirdconnection.FirebirdConnection

registerConnection(['firebird', 'interbase'], builder)



from sqlobject import col, dberrors
from sqlobject.compat import PY2
from sqlobject.converters import registerConverter, StringLikeConverter
from sqlobject.dbconnection import DBAPI


class ErrorMessage(str):
    def __new__(cls, e, append_msg=''):
        if len(e.args) > 1:
            obj = str.__new__(cls, e.args[1] + append_msg)
        else:
            obj = str.__new__(cls, append_msg)
        try:
            obj.code = int(e.args[0])
        except ValueError:
            obj.code = e.args[0]
        obj.module = e.__module__
        obj.exception = e.__class__.__name__
        return obj

mysql_Bin = None


class MySQLConnection(DBAPI):

    supportTransactions = False
    dbName = 'mysql'
    schemes = [dbName]

    odbc_keywords = ('Server', 'Port', 'UID', 'Password', 'Database')

    def __init__(self, db, user, password='', host='localhost', port=0, **kw):
        drivers = kw.pop('driver', None) or 'mysqldb'
        for driver in drivers.split(','):
            driver = driver.strip()
            if not driver:
                continue
            try:
                if driver.lower() in ('mysqldb', 'pymysql'):
                    if driver.lower() == 'pymysql':
                        import pymysql
                        pymysql.install_as_MySQLdb()
                    import MySQLdb
                    if driver.lower() == 'mysqldb':
                        if MySQLdb.version_info[:3] < (1, 2, 2):
                            raise ValueError(
                                'SQLObject requires MySQLdb 1.2.2 or later')
                    import MySQLdb.constants.CR
                    import MySQLdb.constants.ER
                    self.module = MySQLdb
                    if driver.lower() == 'mysqldb':
                        self.CR_SERVER_GONE_ERROR = \
                            MySQLdb.constants.CR.SERVER_GONE_ERROR
                        self.CR_SERVER_LOST = \
                            MySQLdb.constants.CR.SERVER_LOST
                    else:
                        self.CR_SERVER_GONE_ERROR = \
                            MySQLdb.constants.CR.CR_SERVER_GONE_ERROR
                        self.CR_SERVER_LOST = \
                            MySQLdb.constants.CR.CR_SERVER_LOST
                    self.ER_DUP_ENTRY = MySQLdb.constants.ER.DUP_ENTRY
                elif driver in ('connector', 'connector-python'):
                    import mysql.connector
                    self.module = mysql.connector
                    self.CR_SERVER_GONE_ERROR = \
                        mysql.connector.errorcode.CR_SERVER_GONE_ERROR
                    self.CR_SERVER_LOST = \
                        mysql.connector.errorcode.CR_SERVER_LOST
                    self.ER_DUP_ENTRY = mysql.connector.errorcode.ER_DUP_ENTRY
                    if driver == 'connector-python':
                        self.connector_type = 'mysql.connector-python'
                    else:
                        self.connector_type = 'mysql.connector'
                elif driver == 'oursql':
                    import oursql
                    self.module = oursql
                    self.CR_SERVER_GONE_ERROR = \
                        oursql.errnos['CR_SERVER_GONE_ERROR']
                    self.CR_SERVER_LOST = oursql.errnos['CR_SERVER_LOST']
                    self.ER_DUP_ENTRY = oursql.errnos['ER_DUP_ENTRY']
                elif driver == 'mariadb':
                    import mariadb
                    self.module = mariadb
                elif driver == 'pyodbc':
                    import pyodbc
                    self.module = pyodbc
                elif driver == 'pypyodbc':
                    import pypyodbc
                    self.module = pypyodbc
                elif driver == 'odbc':
                    try:
                        import pyodbc
                    except ImportError:
                        import pypyodbc as pyodbc
                    self.module = pyodbc
                else:
                    raise ValueError(
                        'Unknown MySQL driver "%s", '
                        'expected mysqldb, connector, connector-python, '
                        'oursql, pymysql, mariadb, '
                        'odbc, pyodbc or pypyodbc' % driver)
            except ImportError:
                pass
            else:
                break
        else:
            raise ImportError(
                'Cannot find a MySQL driver, tried %s' % drivers)

        self.host = host
        self.port = port or 3306
        self.db = db
        self.user = user
        self.password = password
        self.kw = {}

        for key in ("unix_socket", "init_command",
                    "read_default_file", "read_default_group", "conv"):
            if key in kw:
                self.kw[key] = kw.pop(key)
        for key in ("connect_timeout", "compress", "named_pipe", "use_unicode",
                    "client_flag", "local_infile"):
            if key in kw:
                self.kw[key] = int(kw.pop(key))
        if driver in ('connector', 'connector-python'):
            for key in ("ssl_key", "ssl_cert", "ssl_ca", "ssl_capath"):
                if key in kw:
                    self.kw[key] = kw.pop(key)
        else:
            for key in ("ssl_key", "ssl_cert", "ssl_ca", "ssl_capath"):
                if key in kw:
                    if "ssl" not in self.kw:
                        self.kw["ssl"] = {}
                    self.kw["ssl"][key[4:]] = kw.pop(key)
        if "charset" in kw:
            self.dbEncoding = self.kw["charset"] = kw.pop("charset")
        else:
            self.dbEncoding = None
        self.driver = driver

        if driver in ('mariadb', 'odbc', 'pyodbc', 'pypyodbc'):
            self.CR_SERVER_GONE_ERROR = 2006
            self.CR_SERVER_LOST = 2013
            self.ER_DUP_ENTRY = '23000'

        if driver in ('odbc', 'pyodbc', 'pypyodbc'):
            self.make_odbc_conn_str(kw.pop('odbcdrv',
                                           'MySQL ODBC 5.3 ANSI Driver'),
                                    db, host, port, user, password
                                    )

        elif driver == 'oursql':
            if "use_unicode" not in self.kw:
                self.kw["use_unicode"] = not PY2
            # oursql doesn't implement ping(True) yet
            self.kw["autoreconnect"] = True

        elif driver == 'mariadb':
            self.kw.pop("charset", None)

        elif driver in ('connector', 'connector-python'):
            registerConverter(bytes, ConnectorBytesConverter)

        global mysql_Bin
        if not PY2 and mysql_Bin is None:
            mysql_Bin = self.module.Binary
            self.module.Binary = lambda x: mysql_Bin(x).decode(
                'ascii', errors='surrogateescape')

        self._server_version = None
        self._can_use_microseconds = None
        self._can_use_json_funcs = None

        DBAPI.__init__(self, **kw)

    @classmethod
    def _connectionFromParams(cls, user, password, host, port, path, args):
        return cls(db=path.strip('/'),
                   user=user or '', password=password or '',
                   host=host or 'localhost', port=port, **args)

    def makeConnection(self):
        dbEncoding = self.dbEncoding
        if dbEncoding:
            if self.driver.lower() in ('mysqldb', 'pymysql'):
                from MySQLdb.connections import Connection
                if not hasattr(Connection, 'set_character_set'):
                    # monkeypatch pre MySQLdb 1.2.1
                    def character_set_name(self):
                        return dbEncoding + '_' + dbEncoding
                    Connection.character_set_name = character_set_name
        if self.driver in ('connector', 'connector-python'):
            self.kw['consume_results'] = True
        try:
            if self.driver in ('odbc', 'pyodbc', 'pypyodbc'):
                self.debugWriter.write(
                    "ODBC connect string: " + self.odbc_conn_str)
                conn = self.module.connect(self.odbc_conn_str)
            else:
                conn = self.module.connect(
                    host=self.host, port=self.port, db=self.db,
                    user=self.user, passwd=self.password, **self.kw)
                if self.driver == 'mariadb':
                    # Attempt to reconnect.
                    # This setting is persistent due to ``auto_reconnect``.
                    # mariadb doesn't implement ping(True)
                    conn.auto_reconnect = True
                    conn.ping()
                elif self.driver != 'oursql':
                    # Attempt to reconnect. This setting is persistent.
                    conn.ping(True)
        except self.module.OperationalError as e:
            conninfo = ("; used connection string: "
                        "host=%(host)s, port=%(port)s, "
                        "db=%(db)s, user=%(user)s" % self.__dict__)
            raise dberrors.OperationalError(ErrorMessage(e, conninfo))

        self._setAutoCommit(conn, bool(self.autoCommit))

        if dbEncoding:
            if self.driver in ('odbc', 'pyodbc'):
                conn.setdecoding(self.module.SQL_CHAR, encoding=dbEncoding)
                conn.setdecoding(self.module.SQL_WCHAR, encoding=dbEncoding)
                if PY2:
                    conn.setencoding(str, encoding=dbEncoding)
                    conn.setencoding(unicode, encoding=dbEncoding)  # noqa
                else:
                    conn.setencoding(encoding=dbEncoding)
            elif hasattr(conn, 'set_character_set'):
                conn.set_character_set(dbEncoding)
            elif self.driver == 'oursql':
                conn.charset = dbEncoding
            elif hasattr(conn, 'query'):
                # works along with monkeypatching code above
                conn.query("SET NAMES %s" % dbEncoding)

        return conn

    def _setAutoCommit(self, conn, auto):
        if hasattr(conn, 'autocommit'):
            try:
                conn.autocommit(auto)
            except TypeError:
                # mysql-connector{-python} has autocommit as a property
                conn.autocommit = auto

    def _force_reconnect(self, conn):
        if self.driver.lower() == 'pymysql':
            conn.ping(True)
            self._setAutoCommit(conn, bool(self.autoCommit))
            if self.dbEncoding:
                conn.query("SET NAMES %s" % self.dbEncoding)

    def _executeRetry(self, conn, cursor, query):
        if self.debug:
            self.printDebug(conn, query, 'QueryR')
        dbEncoding = self.dbEncoding
        if dbEncoding and not isinstance(query, bytes) and (
                self.driver in ('mysqldb', 'connector', 'connector-python',
                                'oursql', 'mariadb')):
            query = query.encode(dbEncoding, 'surrogateescape')
        # When a server connection is lost and a query is attempted, most of
        # the time the query will raise a SERVER_LOST exception, then at the
        # second attempt to execute it, the mysql lib will reconnect and
        # succeed. However is a few cases, the first attempt raises the
        # SERVER_GONE exception, the second attempt the SERVER_LOST exception
        # and only the third succeeds. Thus the 3 in the loop count.
        # If it doesn't reconnect even after 3 attempts, while the database is
        # up and running, it is because a 5.0.3 (or newer) server is used
        # which no longer permits autoreconnects by default. In that case a
        # reconnect flag must be set when making the connection to indicate
        # that autoreconnecting is desired. In MySQLdb 1.2.2 or newer this is
        # done by calling ping(True) on the connection.
        # PyMySQL needs explicit reconnect
        # each time we detect connection timeout.
        for count in range(3):
            try:
                return cursor.execute(query)
            except self.module.OperationalError as e:
                if e.args[0] in (self.CR_SERVER_GONE_ERROR,
                                 self.CR_SERVER_LOST):
                    if count == 2:
                        raise dberrors.OperationalError(ErrorMessage(e))
                    if self.debug:
                        self.printDebug(conn, str(e), 'ERROR')
                    if self.driver.lower() == 'pymysql':
                        self._force_reconnect(conn)
                else:
                    raise dberrors.OperationalError(ErrorMessage(e))
            except self.module.IntegrityError as e:
                msg = ErrorMessage(e)
                if e.args[0] == self.ER_DUP_ENTRY:
                    raise dberrors.DuplicateEntryError(msg)
                elif isinstance(e.args[0], str) \
                        and e.args[0].startswith('Duplicate'):
                    raise dberrors.DuplicateEntryError(msg)
                else:
                    raise dberrors.IntegrityError(msg)
            except self.module.InternalError as e:
                raise dberrors.InternalError(ErrorMessage(e))
            except self.module.ProgrammingError as e:
                if e.args[0] is not None:
                    raise dberrors.ProgrammingError(ErrorMessage(e))
            except self.module.DataError as e:
                raise dberrors.DataError(ErrorMessage(e))
            except self.module.NotSupportedError as e:
                raise dberrors.NotSupportedError(ErrorMessage(e))
            except self.module.DatabaseError as e:
                raise dberrors.DatabaseError(ErrorMessage(e))
            except self.module.InterfaceError as e:
                raise dberrors.InterfaceError(ErrorMessage(e))
            except self.module.Warning as e:
                raise Warning(ErrorMessage(e))
            except self.module.Error as e:
                raise dberrors.Error(ErrorMessage(e))

    def _queryInsertID(self, conn, soInstance, id, names, values):
        table = soInstance.sqlmeta.table
        idName = soInstance.sqlmeta.idName
        c = conn.cursor()
        if id is not None:
            names = [idName] + names
            values = [id] + values
        q = self._insertSQL(table, names, values)
        if self.debug:
            self.printDebug(conn, q, 'QueryIns')
        self._executeRetry(conn, c, q)
        if id is None:
            try:
                id = c.lastrowid
            except AttributeError:
                try:
                    id = c.insert_id
                except AttributeError:
                    self._executeRetry(conn, c, "SELECT LAST_INSERT_ID();")
                    id = c.fetchone()[0]
                else:
                    id = c.insert_id()
        c.close()
        if self.debugOutput:
            self.printDebug(conn, id, 'QueryIns', 'result')
        return id

    @classmethod
    def _queryAddLimitOffset(cls, query, start, end):
        if not start:
            return "%s LIMIT %i" % (query, end)
        if not end:
            return "%s LIMIT %i, -1" % (query, start)
        return "%s LIMIT %i, %i" % (query, start, end - start)

    def createReferenceConstraint(self, soClass, col):
        return col.mysqlCreateReferenceConstraint()

    def createColumn(self, soClass, col):
        return col.mysqlCreateSQL(self)

    def createIndexSQL(self, soClass, index):
        return index.mysqlCreateIndexSQL(soClass)

    def createIDColumn(self, soClass):
        if soClass.sqlmeta.idType is str:
            return '%s TEXT PRIMARY KEY' % soClass.sqlmeta.idName
        if soClass.sqlmeta.idType is not int:
            raise TypeError('sqlmeta.idType must be int or str, not %r'
                            % soClass.sqlmeta.idType)
        if soClass.sqlmeta.idSize is None:
            mysql_int_type = 'INT'
        elif soClass.sqlmeta.idSize in ('TINY', 'SMALL', 'MEDIUM', 'BIG'):
            mysql_int_type = '%sINT' % soClass.sqlmeta.idSize
        else:
            raise ValueError(
                "sqlmeta.idSize must be 'TINY', 'SMALL', 'MEDIUM', 'BIG' "
                "or None, not %r" % soClass.sqlmeta.idSize)
        return '%s %s PRIMARY KEY AUTO_INCREMENT' \
               % (soClass.sqlmeta.idName, mysql_int_type)

    def joinSQLType(self, join):
        return 'INT NOT NULL'

    def tableExists(self, tableName):
        try:
            # Use DESCRIBE instead of SHOW TABLES because SHOW TABLES
            # assumes there is a default database selected
            # which is not always True (for an embedded application, e.g.)
            self.query('DESCRIBE %s' % (tableName))
            return True
        except dberrors.ProgrammingError as e:
            if e.args[0].code in (1146, '42S02'):  # ER_NO_SUCH_TABLE
                return False
            if self.driver == 'mariadb':
                return False
            raise

    def addColumn(self, tableName, column):
        self.query('ALTER TABLE %s ADD COLUMN %s' %
                   (tableName,
                    column.mysqlCreateSQL(self)))

    def delColumn(self, sqlmeta, column):
        self.query('ALTER TABLE %s DROP COLUMN %s' % (sqlmeta.table,
                                                      column.dbName))

    def columnsFromSchema(self, tableName, soClass):
        colData = self.queryAll("SHOW COLUMNS FROM %s"
                                % tableName)
        results = []
        for field, t, nullAllowed, key, default, extra in colData:
            if field == soClass.sqlmeta.idName:
                continue
            colClass, kw = self.guessClass(t)
            if self.kw.get('use_unicode') and colClass is col.StringCol:
                colClass = col.UnicodeCol
                if self.dbEncoding:
                    kw['dbEncoding'] = self.dbEncoding
            kw['name'] = soClass.sqlmeta.style.dbColumnToPythonAttr(field)
            kw['dbName'] = field

            # Since MySQL 5.0, 'NO' is returned in the NULL column
            # (SQLObject expected '')
            kw['notNone'] = (nullAllowed.upper() != 'YES' and True or False)

            if not PY2 and isinstance(t, bytes):
                t = t.decode('ascii')

            if default and t.startswith('int'):
                kw['default'] = int(default)
            elif default and t.startswith('float'):
                kw['default'] = float(default)
            elif default == 'CURRENT_TIMESTAMP' and t == 'timestamp':
                kw['default'] = None
            elif default and colClass is col.BoolCol:
                kw['default'] = int(default) and True or False
            else:
                kw['default'] = default
            # @@ skip key...
            # @@ skip extra...
            results.append(colClass(**kw))
        return results

    def guessClass(self, t):
        if not PY2 and isinstance(t, bytes):
            t = t.decode('ascii')
        if t.startswith('int'):
            return col.IntCol, {}
        elif t.startswith('enum'):
            values = []
            for i in t[5:-1].split(','):  # take the enum() off and split
                values.append(i[1:-1])  # remove the surrounding \'
            return col.EnumCol, {'enumValues': values}
        elif t.startswith('double'):
            return col.FloatCol, {}
        elif t.startswith('varchar'):
            colType = col.StringCol
            if self.kw.get('use_unicode', False):
                colType = col.UnicodeCol
            if t.endswith('binary'):
                return colType, {'length': int(t[8:-8]),
                                 'char_binary': True}
            else:
                return colType, {'length': int(t[8:-1])}
        elif t.startswith('char'):
            if t.endswith('binary'):
                return col.StringCol, {'length': int(t[5:-8]),
                                       'varchar': False,
                                       'char_binary': True}
            else:
                return col.StringCol, {'length': int(t[5:-1]),
                                       'varchar': False}
        elif t.startswith('datetime'):
            return col.DateTimeCol, {}
        elif t.startswith('date'):
            return col.DateCol, {}
        elif t.startswith('time'):
            return col.TimeCol, {}
        elif t.startswith('timestamp'):
            return col.TimestampCol, {}
        elif t.startswith('bool'):
            return col.BoolCol, {}
        elif t.startswith('tinyblob'):
            return col.BLOBCol, {"length": 2 ** 8 - 1}
        elif t.startswith('tinytext'):
            return col.StringCol, {"length": 2 ** 8 - 1, "varchar": True}
        elif t.startswith('blob'):
            return col.BLOBCol, {"length": 2 ** 16 - 1}
        elif t.startswith('text'):
            return col.StringCol, {"length": 2 ** 16 - 1, "varchar": True}
        elif t.startswith('mediumblob'):
            return col.BLOBCol, {"length": 2 ** 24 - 1}
        elif t.startswith('mediumtext'):
            return col.StringCol, {"length": 2 ** 24 - 1, "varchar": True}
        elif t.startswith('longblob'):
            return col.BLOBCol, {"length": 2 ** 32}
        elif t.startswith('longtext'):
            return col.StringCol, {"length": 2 ** 32, "varchar": True}
        else:
            return col.Col, {}

    def listTables(self):
        return _decodeBytearrays(self.queryAll("SHOW TABLES"))

    def listDatabases(self):
        return _decodeBytearrays(self.queryAll("SHOW DATABASES"))

    def _createOrDropDatabase(self, op="CREATE"):
        self.query('%s DATABASE %s' % (op, self.db))

    def createEmptyDatabase(self):
        self._createOrDropDatabase()

    def dropDatabase(self):
        self._createOrDropDatabase(op="DROP")

    def server_version(self):
        if self._server_version is not None:
            return self._server_version
        try:
            server_version = self.queryOne("SELECT VERSION()")[0]
            server_version = server_version.split('-', 1)
            db_tag = "MySQL"
            if len(server_version) == 2:
                if "MariaDB" in server_version[1]:
                    db_tag = "MariaDB"
                server_version = server_version[0]
            server_version = tuple(int(v) for v in server_version.split('.'))
            server_version = (server_version, db_tag)
        except Exception:
            server_version = None  # unknown
        self._server_version = server_version
        return server_version

    def can_use_microseconds(self):
        if self._can_use_microseconds is not None:
            return self._can_use_microseconds
        server_version = self.server_version()
        if server_version is None:
            return None
        server_version, db_tag = server_version
        if db_tag == "MariaDB":
            can_use_microseconds = (server_version >= (5, 3, 0))
        else:  # MySQL
            can_use_microseconds = (server_version >= (5, 6, 4))
        self._can_use_microseconds = can_use_microseconds
        return can_use_microseconds

    def can_use_json_funcs(self):
        if self._can_use_json_funcs is not None:
            return self._can_use_json_funcs
        server_version = self.server_version()
        if server_version is None:
            return None
        server_version, db_tag = server_version
        if db_tag == "MariaDB":
            can_use_json_funcs = (server_version >= (10, 2, 7))
        else:  # MySQL
            can_use_json_funcs = (server_version >= (5, 7, 0))
        self._can_use_json_funcs = can_use_json_funcs
        return can_use_json_funcs


def ConnectorBytesConverter(value, db):
    if not PY2:
        # For PY2 this converter is called also for SQLite
        assert db == 'mysql'
        value = value.decode('latin1')
    return StringLikeConverter(value, db)


def _decodeBytearrays(v_list):
    if not v_list:
        return []
    if not PY2 and isinstance(v_list[0][0], bytearray):
        return [v[0].decode('ascii') for v in v_list]
    return [v[0] for v in v_list]



import csv
from datetime import datetime
from sqlobject.util.csvimport import load_csv

csv_data = """\
name:str,age:datetime,value:int
Test,2000-01-01 21:44:33,42"""


def test_load_csv():
    loaded = load_csv(csv.reader(csv_data.split('\n')),
                      default_class='SQLObject')
    assert loaded == {
        'SQLObject': [
            {
                'age': datetime(2000, 1, 1, 21, 44, 33),
                'name': 'Test',
                'value': 42
            }
        ]
    }



from sqlobject import ForeignKey, IntCol, SQLObject, StringCol
from sqlobject.inheritance import InheritableSQLObject
from sqlobject.versioning import Versioning
from sqlobject.tests.dbtest import setupClass


class MyClass(SQLObject):
    name = StringCol()
    versions = Versioning()


class Base(InheritableSQLObject):
    name = StringCol()
    so_value = IntCol(default=0)
    versions = Versioning()


class Child(Base):
    toy = StringCol()


class Government(InheritableSQLObject):
    name = StringCol()


class Monarchy(Government):
    monarch = StringCol()
    versions = Versioning()


class VChild(Base):
    weapon = StringCol()
    versions = Versioning()


class HasForeign(SQLObject):
    foreign = ForeignKey("Base")
    versions = Versioning()


def _set_extra():
    return "read all about it"


class Extra(SQLObject):
    name = StringCol()
    versions = Versioning(
        extraCols={'extra': StringCol(default=_set_extra())})


class HasAltId(SQLObject):
    name = StringCol()
    altid = IntCol(alternateID=True)
    versions = Versioning()


def setup():
    classes = [MyClass, Base, Child, Government,
               Monarchy, VChild, Extra, HasAltId]
    if hasattr(HasForeign, "_connection"):
        classes.insert(0, HasForeign)
    else:
        classes.append(HasForeign)
    for cls in classes:
        if hasattr(cls, 'versions') and getattr(cls, "_connection", None) and \
                cls._connection.tableExists(cls.sqlmeta.table):
            setupClass(cls.versions.versionClass)
        setupClass(cls)
        if hasattr(cls, 'versions'):
            setupClass(cls.versions.versionClass)
            for version in cls.versions.versionClass.select():
                version.destroySelf()


def test_versioning():
    # the simple case
    setup()
    mc = MyClass(name='fleem')
    mc.set(name='morx')
    assert len(list(mc.versions)) == 1
    assert mc.versions[0].name == "fleem"

    assert len(list(MyClass.select())) == 1


def test_inheritable_versioning():
    setup()

    # base versioned, child unversioned
    base = Base(name='fleem')
    base.set(name='morx')
    assert len(list(base.versions)) == 1
    assert base.versions[0].name == "fleem"
    assert len(list(Base.select())) == 1

    child = Child(name='child', toy='nintendo')
    child.set(name='teenager', toy='guitar')
    assert len(list(child.versions)) == 0

    # child versioned, base unversioned
    government = Government(name='canada')
    assert not hasattr(government, 'versions')

    monarchy = Monarchy(name='UK', monarch='king george iv')
    assert len(list(monarchy.versions)) == 0
    monarchy.set(name='queen elisabeth ii')
    assert len(list(monarchy.versions)) == 1
    assert monarchy.versions[0].name == "UK"
    assert len(list(Monarchy.select())) == 1

    # both parent and child versioned
    num_base_versions = len(list(base.versions))
    vchild = VChild(name='kid', weapon='slingshot')
    vchild.set(name='toon', weapon='dynamite')
    assert len(list(base.versions)) == num_base_versions
    assert len(list(vchild.versions)) == 1
    # test setting using setattr directly rather than .set
    vchild.name = "newname"
    assert len(list(vchild.versions)) == 2


def test_restore():
    setup()
    base = Base(name='fleem')
    base.set(name='morx')
    assert base.name == "morx"
    base.versions[0].restore()
    assert base.name == "fleem"

    monarchy = Monarchy(name='USA', monarch='Emperor Norton I')
    monarchy.set(name='morx')
    assert monarchy.name == "morx"
    monarchy.versions[0].restore()
    assert monarchy.name == "USA"
    assert monarchy.monarch == "Emperor Norton I"

    extra = Extra(name='fleem')
    extra.set(name='morx')
    assert extra.name == "morx"
    extra.versions[0].restore()
    assert extra.name == "fleem"


def test_next():
    setup()
    base = Base(name='first', so_value=1)
    base.set(name='second')
    base.set(name='third', so_value=2)
    version = base.versions[0]
    assert version.nextVersion() == base.versions[1]
    assert version.nextVersion().nextVersion() == base


def test_get_changed():
    setup()
    base = Base(name='first', so_value=1)
    base.set(name='second')
    base.set(name='third', so_value=2)
    assert base.versions[0].getChangedFields() == ['Name']
    assert sorted(base.versions[1].getChangedFields()) == ['Name', 'So_Value']


def test_foreign_keys():
    setup()
    base1 = Base(name='first', so_value=1)
    base2 = Base(name='first', so_value=1)
    has_foreign = HasForeign(foreign=base1)
    has_foreign.foreign = base2
    assert has_foreign.versions[0].foreign == base1


def test_extra():
    setup()
    extra = Extra(name='title')
    extra.name = 'new'
    assert extra.versions[0].extra == 'read all about it'
    assert sorted(extra.versions[0].getChangedFields()) == ['Name']


def test_altid():
    setup()
    extra = HasAltId(name="fleem", altid=5)
    extra.name = "morx"



"""
Bound attributes are attributes that are bound to a specific class and
a specific name.  In SQLObject a typical example is a column object,
which knows its name and class.

A bound attribute should define a method ``__addtoclass__(added_class,
name)`` (attributes without this method will simply be treated as
normal).  The return value is ignored; if the attribute wishes to
change the value in the class, it must call ``setattr(added_class,
name, new_value)``.

BoundAttribute is a class that facilitates lazy attribute creation.
"""
from __future__ import absolute_import

from . import declarative
from . import events

__all__ = ['BoundAttribute', 'BoundFactory']


class BoundAttribute(declarative.Declarative):

    """
    This is a declarative class that passes all the values given to it
    to another object.  So you can pass it arguments (via
    __init__/__call__) or give it the equivalent of keyword arguments
    through subclassing.  Then a bound object will be added in its
    place.

    To hook this other object in, override ``make_object(added_class,
    name, **attrs)`` and maybe ``set_object(added_class, name,
    **attrs)`` (the default implementation of ``set_object``
    just resets the attribute to whatever ``make_object`` returned).

    Also see ``BoundFactory``.
    """

    _private_variables = (
        '_private_variables',
        '_all_attributes',
        '__classinit__',
        '__addtoclass__',
        '_add_attrs',
        'set_object',
        'make_object',
        'clone_in_subclass',
    )

    _all_attrs = ()
    clone_for_subclass = True

    def __classinit__(cls, new_attrs):
        declarative.Declarative.__classinit__(cls, new_attrs)
        cls._all_attrs = cls._add_attrs(cls, new_attrs)

    def __instanceinit__(self, new_attrs):
        declarative.Declarative.__instanceinit__(self, new_attrs)
        self.__dict__['_all_attrs'] = self._add_attrs(self, new_attrs)

    @staticmethod
    def _add_attrs(this_object, new_attrs):
        private = this_object._private_variables
        all_attrs = list(this_object._all_attrs)
        for key in new_attrs.keys():
            if key.startswith('_') or key in private:
                continue
            if key not in all_attrs:
                all_attrs.append(key)
        return tuple(all_attrs)

    @declarative.classinstancemethod
    def __addtoclass__(self, cls, added_class, attr_name):
        me = self or cls
        attrs = {}
        for name in me._all_attrs:
            attrs[name] = getattr(me, name)
        attrs['added_class'] = added_class
        attrs['attr_name'] = attr_name
        obj = me.make_object(**attrs)

        if self.clone_for_subclass:
            def on_rebind(new_class_name, bases, new_attrs,
                          post_funcs, early_funcs):
                def rebind(new_class):
                    me.set_object(
                        new_class, attr_name,
                        me.make_object(**attrs))
                post_funcs.append(rebind)
            events.listen(receiver=on_rebind, soClass=added_class,
                          signal=events.ClassCreateSignal, weak=False)

        me.set_object(added_class, attr_name, obj)

    @classmethod
    def set_object(cls, added_class, attr_name, obj):
        setattr(added_class, attr_name, obj)

    @classmethod
    def make_object(cls, added_class, attr_name, *args, **attrs):
        raise NotImplementedError

    def __setattr__(self, name, value):
        self.__dict__['_all_attrs'] = self._add_attrs(self, {name: value})
        self.__dict__[name] = value


class BoundFactory(BoundAttribute):

    """
    This will bind the attribute to whatever is given by
    ``factory_class``.  This factory should be a callable with the
    signature ``factory_class(added_class, attr_name, *args, **kw)``.

    The factory will be reinvoked (and the attribute rebound) for
    every subclassing.
    """

    factory_class = None
    _private_variables = (
        BoundAttribute._private_variables + ('factory_class',))

    def make_object(cls, added_class, attr_name, *args, **kw):
        return cls.factory_class(added_class, attr_name, *args, **kw)



import math
import threading
import pytest
from sqlobject import SQLObject, StringCol, FloatCol
from sqlobject.compat import string_type
from sqlobject.sqlbuilder import Select
from sqlobject.tests.dbtest import getConnection, setupClass, supports
from sqlobject.tests.dbtest import setSQLiteConnectionFactory
from .test_basic import SOTestSO1


try:
    connection = getConnection()
except (AttributeError, NameError):
    # The module was imported during documentation building
    pass
else:
    if connection.dbName != "sqlite":
        pytestmark = pytest.mark.skip("These tests require SQLite")


class SQLiteFactoryTest(SQLObject):
    name = StringCol()


def test_sqlite_factory():
    setupClass(SQLiteFactoryTest)

    if not SQLiteFactoryTest._connection.using_sqlite2:
        pytest.skip("These tests require SQLite v2+")

    factory = [None]

    def SQLiteConnectionFactory(sqlite):
        class MyConnection(sqlite.Connection):
            pass
        factory[0] = MyConnection
        return MyConnection

    setSQLiteConnectionFactory(SQLiteFactoryTest, SQLiteConnectionFactory)

    conn = SQLiteFactoryTest._connection.makeConnection()
    assert factory[0]
    assert isinstance(conn, factory[0])


def test_sqlite_factory_str():
    setupClass(SQLiteFactoryTest)

    if not SQLiteFactoryTest._connection.using_sqlite2:
        pytest.skip("These tests require SQLite v2+")

    factory = [None]

    def SQLiteConnectionFactory(sqlite):
        class MyConnection(sqlite.Connection):
            pass
        factory[0] = MyConnection
        return MyConnection

    from sqlobject.sqlite import sqliteconnection
    sqliteconnection.SQLiteConnectionFactory = SQLiteConnectionFactory

    setSQLiteConnectionFactory(SQLiteFactoryTest, "SQLiteConnectionFactory")

    conn = SQLiteFactoryTest._connection.makeConnection()
    assert factory[0]
    assert isinstance(conn, factory[0])
    del sqliteconnection.SQLiteConnectionFactory


def test_sqlite_aggregate():
    setupClass(SQLiteFactoryTest)

    if not SQLiteFactoryTest._connection.using_sqlite2:
        pytest.skip("These tests require SQLite v2+")

    def SQLiteConnectionFactory(sqlite):
        class MyConnection(sqlite.Connection):
            def __init__(self, *args, **kwargs):
                super(MyConnection, self).__init__(*args, **kwargs)
                self.create_aggregate("group_concat", 1, self.group_concat)

            class group_concat:
                def __init__(self):
                    self.acc = []

                def step(self, value):
                    if isinstance(value, string_type):
                        self.acc.append(value)
                    else:
                        self.acc.append(str(value))

                def finalize(self):
                    self.acc.sort()
                    return ", ".join(self.acc)

        return MyConnection

    setSQLiteConnectionFactory(SQLiteFactoryTest, SQLiteConnectionFactory)

    SQLiteFactoryTest(name='sqlobject')
    SQLiteFactoryTest(name='sqlbuilder')
    assert SQLiteFactoryTest.select(orderBy="name").\
        accumulateOne("group_concat", "name") == \
        "sqlbuilder, sqlobject"


def do_select():
    list(SOTestSO1.select())


def test_sqlite_threaded():
    setupClass(SOTestSO1)
    t = threading.Thread(target=do_select)
    t.start()
    t.join()
    # This should reuse the same connection as the connection
    # made above (at least will with most database drivers, but
    # this will cause an error in SQLite):
    do_select()


def test_empty_string():
    setupClass(SOTestSO1)
    test = SOTestSO1(name=None, passwd='')
    assert test.name is None
    assert test.passwd == ''


def test_memorydb():
    if not supports("memorydb"):
        pytest.skip("memorydb isn't supported")
    if not connection._memory:
        pytest.skip("The connection isn't memorydb")
    setupClass(SOTestSO1)
    connection.close()  # create a new connection to an in-memory database
    SOTestSO1.setConnection(connection)
    SOTestSO1.createTable()


def test_list_databases():
    assert 'main' in connection.listDatabases()


def test_list_tables():
    setupClass(SOTestSO1)
    assert SOTestSO1.sqlmeta.table in connection.listTables()


class SQLiteTruedivTest(SQLObject):
    value = FloatCol()


def test_truediv():
    setupClass(SQLiteTruedivTest)

    if SQLiteTruedivTest._connection.dbName == "sqlite":
        if not SQLiteTruedivTest._connection.using_sqlite2:
            pytest.skip("These tests require SQLite v2+")

        def SQLiteConnectionFactory(sqlite):
            class MyConnection(sqlite.Connection):
                def __init__(self, *args, **kwargs):
                    super(MyConnection, self).__init__(*args, **kwargs)
                    self.create_function("floor", 1, math.floor)
            return MyConnection

        setSQLiteConnectionFactory(SQLiteTruedivTest, SQLiteConnectionFactory)

    SQLiteTruedivTest(value=-5.0)
    assert SQLiteTruedivTest._connection.queryAll(
        SQLiteTruedivTest._connection.sqlrepr(
            Select(SQLiteTruedivTest.q.value // 4)))[0][0] == -2



from .main import SQLObject
from .sqlbuilder import AND, Alias, ColumnAS, LEFTJOINOn, \
    NoDefault, SQLCall, SQLConstant, SQLObjectField, SQLObjectTable, SQLOp, \
    Select, sqlrepr


class ViewSQLObjectField(SQLObjectField):
    def __init__(self, alias, *arg):
        SQLObjectField.__init__(self, *arg)
        self.alias = alias

    def __sqlrepr__(self, db):
        return self.alias + "." + self.fieldName

    def tablesUsedImmediate(self):
        return [self.tableName]


class ViewSQLObjectTable(SQLObjectTable):
    FieldClass = ViewSQLObjectField

    def __getattr__(self, attr):
        if attr == 'sqlmeta':
            raise AttributeError
        return SQLObjectTable.__getattr__(self, attr)

    def _getattrFromID(self, attr):
        return self.FieldClass(self.soClass.sqlmeta.alias, self.tableName,
                               'id', attr, self.soClass, None)

    def _getattrFromColumn(self, column, attr):
        return self.FieldClass(self.soClass.sqlmeta.alias, self.tableName,
                               column.name, attr, self.soClass, column)


class ViewSQLObject(SQLObject):
    """
    A SQLObject class that derives all it's values from other SQLObject
    classes. Columns on subclasses should use SQLBuilder constructs for dbName,
    and sqlmeta should specify:

    * idName as a SQLBuilder construction
    * clause as SQLBuilder clause for specifying join conditions
      or other restrictions
    * table as an optional alternate name for the class alias

    See test_views.py for simple examples.
    """

    def __classinit__(cls, new_attrs):
        SQLObject.__classinit__(cls, new_attrs)
        # like is_base
        if cls.__name__ != 'ViewSQLObject':
            dbName = hasattr(cls, '_connection') and \
                (cls._connection and cls._connection.dbName) or None

            if getattr(cls.sqlmeta, 'table', None):
                cls.sqlmeta.alias = cls.sqlmeta.table
            else:
                cls.sqlmeta.alias = \
                    cls.sqlmeta.style.pythonClassToDBTable(cls.__name__)
            alias = cls.sqlmeta.alias
            columns = [ColumnAS(cls.sqlmeta.idName, 'id')]
            # {sqlrepr-key: [restriction, *aggregate-column]}
            aggregates = {'': [None]}
            inverseColumns = dict(
                [(y, x) for x, y in cls.sqlmeta.columns.items()])
            for col in cls.sqlmeta.columnList:
                n = inverseColumns[col]
                ascol = ColumnAS(col.dbName, n)
                if isAggregate(col.dbName):
                    restriction = getattr(col, 'aggregateClause', None)
                    if restriction:
                        restrictkey = sqlrepr(restriction, dbName)
                        aggregates[restrictkey] = \
                            aggregates.get(restrictkey, [restriction]) + \
                            [ascol]
                    else:
                        aggregates[''].append(ascol)
                else:
                    columns.append(ascol)

            metajoin = getattr(cls.sqlmeta, 'join', NoDefault)
            clause = getattr(cls.sqlmeta, 'clause', NoDefault)
            select = Select(columns,
                            distinct=True,
                            # @@ LDO check if this really mattered
                            # for performance
                            # @@ Postgres (and MySQL?) extension!
                            # distinctOn=cls.sqlmeta.idName,
                            join=metajoin,
                            clause=clause)

            aggregates = aggregates.values()

            if aggregates != [[None]]:
                join = []
                last_alias = "%s_base" % alias
                last_id = "id"
                last = Alias(select, last_alias)
                columns = [
                    ColumnAS(SQLConstant("%s.%s" % (last_alias, x.expr2)),
                             x.expr2) for x in columns]

                for i, agg in enumerate(aggregates):
                    restriction = agg[0]
                    if restriction is None:
                        restriction = clause
                    else:
                        restriction = AND(clause, restriction)
                    agg = agg[1:]
                    agg_alias = "%s_%s" % (alias, i)
                    agg_id = '%s_id' % agg_alias
                    if not last.q.alias.endswith('base'):
                        last = None
                    new_alias = Alias(Select(
                        [ColumnAS(cls.sqlmeta.idName, agg_id)] + agg,
                        groupBy=cls.sqlmeta.idName,
                        join=metajoin,
                        clause=restriction),
                        agg_alias)
                    agg_join = LEFTJOINOn(last, new_alias,
                                          "%s.%s = %s.%s" % (
                                              last_alias, last_id,
                                              agg_alias, agg_id))

                    join.append(agg_join)
                    for col in agg:
                        columns.append(
                            ColumnAS(SQLConstant(
                                "%s.%s" % (agg_alias, col.expr2)),
                                col.expr2))

                    last = new_alias
                    last_alias = agg_alias
                    last_id = agg_id
                select = Select(columns,
                                join=join)

            cls.sqlmeta.table = Alias(select, alias)
            cls.q = ViewSQLObjectTable(cls)
            for n, col in cls.sqlmeta.columns.items():
                col.dbName = n


def isAggregate(expr):
    if isinstance(expr, SQLCall):
        return True
    if isinstance(expr, SQLOp):
        return isAggregate(expr.expr1) or isAggregate(expr.expr2)
    return False



from __future__ import print_function
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from sqlobject import IntCol, SQLObject, StringCol
from sqlobject.util.csvexport import export_csv, export_csv_zip
from .dbtest import setupClass


def assert_export(result, *args, **kw):
    f = StringIO()
    kw['writer'] = f
    export_csv(*args, **kw)
    s = f.getvalue().replace('\r\n', '\n')
    if result.strip() != s.strip():
        print('**Expected:')
        print(result)
        print('**Got:')
        print(s)
        assert result.strip() == s.strip()


class SimpleCSV(SQLObject):

    name = StringCol()
    address = StringCol()
    address.csvTitle = 'Street Address'
    hidden = StringCol()
    hidden.noCSV = True


class ComplexCSV(SQLObject):

    fname = StringCol()
    lname = StringCol()
    age = IntCol()
    extraCSVColumns = [('name', 'Full Name'), 'initials']
    # initials should end up at the end then:
    csvColumnOrder = ['name', 'fname', 'lname', 'age']

    def _get_name(self):
        return self.fname + ' ' + self.lname

    def _get_initials(self):
        return self.fname[0] + self.lname[0]


def test_simple():
    setupClass(SimpleCSV)
    SimpleCSV(name='Bob', address='432W', hidden='boo')
    SimpleCSV(name='Joe', address='123W', hidden='arg')
    assert_export("""\
name,Street Address
Bob,432W
Joe,123W
""", SimpleCSV, orderBy='name')
    assert_export("""\
name,Street Address
Joe,123W
Bob,432W
""", SimpleCSV, orderBy='address')
    assert_export("""\
name,Street Address
Joe,123W
""", SimpleCSV.selectBy(name='Joe'))


def test_complex():
    setupClass(ComplexCSV)
    ComplexCSV(fname='John', lname='Doe', age=40)
    ComplexCSV(fname='Bob', lname='Dylan', age=60)
    ComplexCSV(fname='Harriet', lname='Tubman', age=160)
    assert_export("""\
Full Name,fname,lname,age,initials
John Doe,John,Doe,40,JD
Bob Dylan,Bob,Dylan,60,BD
Harriet Tubman,Harriet,Tubman,160,HT
""", ComplexCSV, orderBy='lname')
    assert_export("""\
Full Name,fname,lname,age,initials
Bob Dylan,Bob,Dylan,60,BD
John Doe,John,Doe,40,JD
""", ComplexCSV.select(ComplexCSV.q.lname.startswith('D'), orderBy='fname'))


def test_zip():
    # Just exercise tests, doesn't actually test results
    setupClass(SimpleCSV)
    SimpleCSV(name='Bob', address='432W', hidden='boo')
    SimpleCSV(name='Joe', address='123W', hidden='arg')

    setupClass(ComplexCSV)
    ComplexCSV(fname='John', lname='Doe', age=40)
    ComplexCSV(fname='Bob', lname='Dylan', age=60)
    ComplexCSV(fname='Harriet', lname='Tubman', age=160)
    s = export_csv_zip([SimpleCSV, ComplexCSV])
    assert isinstance(s, bytes) and s
    s = export_csv_zip([SimpleCSV.selectBy(name='Bob'),
                        (ComplexCSV, list(ComplexCSV.selectBy(fname='John')))])



from sqlobject import ForeignKey, SQLObject, SingleJoin, StringCol
from sqlobject.tests.dbtest import setupClass


class PersonWithAlbum(SQLObject):
    name = StringCol()
    # albumNone returns the album or none
    albumNone = SingleJoin('PhotoAlbum', joinColumn='test_person_id')
    # albumInstance returns the album or an default album instance
    albumInstance = SingleJoin('PhotoAlbum', makeDefault=True,
                               joinColumn='test_person_id')


class PhotoAlbum(SQLObject):
    color = StringCol(default='red')
    person = ForeignKey('PersonWithAlbum', dbName='test_person_id')


def test_1():
    setupClass([PersonWithAlbum, PhotoAlbum])

    person = PersonWithAlbum(name='Gokou (Kakarouto)')
    # We didn't create an album, this way it returns None
    assert not person.albumNone
    assert isinstance(person.albumInstance, PhotoAlbum)

    PhotoAlbum(person=person)
    assert person.albumNone
    assert isinstance(person.albumNone, PhotoAlbum)
    assert isinstance(person.albumInstance, PhotoAlbum)



from datetime import datetime, date, time
import pytest

from sqlobject import SQLObject
from sqlobject import col
from sqlobject.col import DateCol, DateTimeCol, TimeCol, use_microseconds, \
    DATETIME_IMPLEMENTATION, MXDATETIME_IMPLEMENTATION, mxdatetime_available, \
    ZOPE_DATETIME_IMPLEMENTATION, zope_datetime_available
from sqlobject.tests.dbtest import getConnection, setupClass
from sqlobject.converters import pendulumDateTimeType

########################################
# Date/time columns
########################################


col.default_datetime_implementation = DATETIME_IMPLEMENTATION


class DateTime1(SQLObject):
    col1 = DateTimeCol()
    col2 = DateCol()
    col3 = TimeCol()


def test_dateTime():
    setupClass(DateTime1)
    _now = datetime.now().replace(microsecond=0)
    dt1 = DateTime1(col1=_now, col2=_now, col3=_now.time())

    assert isinstance(dt1.col1, datetime)
    assert dt1.col1.year == _now.year
    assert dt1.col1.month == _now.month
    assert dt1.col1.day == _now.day
    assert dt1.col1.hour == _now.hour
    assert dt1.col1.minute == _now.minute
    assert dt1.col1.second == _now.second

    assert isinstance(dt1.col2, date)
    assert not isinstance(dt1.col2, datetime)
    assert dt1.col2.year == _now.year
    assert dt1.col2.month == _now.month
    assert dt1.col2.day == _now.day

    assert isinstance(dt1.col3, time)
    assert dt1.col3.hour == _now.hour
    assert dt1.col3.minute == _now.minute
    assert dt1.col3.second == _now.second


def test_microseconds():
    connection = getConnection()
    if not hasattr(connection, 'can_use_microseconds') or \
            not connection.can_use_microseconds():
        pytest.skip(
            "The database doesn't support microseconds; "
            "microseconds are supported by MariaDB since version 5.3.0, "
            "by MySQL since version 5.6.4, "
            "by MSSQL since MS SQL Server 2008.")

    setupClass(DateTime1)
    _now = datetime.now()
    dt1 = DateTime1(col1=_now, col2=_now, col3=_now.time())

    assert dt1.col1.microsecond == _now.microsecond
    assert dt1.col3.microsecond == _now.microsecond

    use_microseconds(False)
    setupClass(DateTime1)
    _now = datetime.now()
    dt1 = DateTime1(col1=_now, col2=_now, col3=_now.time())

    assert dt1.col1.microsecond == 0
    assert dt1.col3.microsecond == 0

    use_microseconds(True)
    setupClass(DateTime1)
    _now = datetime.now()
    dt1 = DateTime1(col1=_now, col2=_now, col3=_now.time())

    assert dt1.col1.microsecond == _now.microsecond
    assert dt1.col3.microsecond == _now.microsecond

if mxdatetime_available:
    col.default_datetime_implementation = MXDATETIME_IMPLEMENTATION
    from mx.DateTime import now as mx_now, Time as mxTime

    dateFormat = None  # use default
    try:
        connection = getConnection()
    except AttributeError:
        # The module was imported during documentation building
        pass
    else:
        if connection.dbName == "sqlite":
            if connection.using_sqlite2:
                # mxDateTime sends and PySQLite2 returns
                # full date/time for dates
                dateFormat = "%Y-%m-%d %H:%M:%S.%f"

    class DateTime2(SQLObject):
        col1 = DateTimeCol()
        col2 = DateCol(dateFormat=dateFormat)
        col3 = TimeCol()

    def test_mxDateTime():
        setupClass(DateTime2)
        _now = mx_now()
        dt2 = DateTime2(col1=_now, col2=_now.pydate(),
                        col3=mxTime(_now.hour, _now.minute, _now.second))

        assert isinstance(dt2.col1, col.mxDateTimeType)
        assert dt2.col1.year == _now.year
        assert dt2.col1.month == _now.month
        assert dt2.col1.day == _now.day
        assert dt2.col1.hour == _now.hour
        assert dt2.col1.minute == _now.minute
        assert dt2.col1.second == int(_now.second)

        assert isinstance(dt2.col2, col.mxDateTimeType)
        assert dt2.col2.year == _now.year
        assert dt2.col2.month == _now.month
        assert dt2.col2.day == _now.day
        assert dt2.col2.hour == 0
        assert dt2.col2.minute == 0
        assert dt2.col2.second == 0

        assert isinstance(dt2.col3, (col.mxDateTimeType, col.mxTimeType))
        assert dt2.col3.hour == _now.hour
        assert dt2.col3.minute == _now.minute
        assert dt2.col3.second == int(_now.second)

if pendulumDateTimeType:
    col.default_datetime_implementation = DATETIME_IMPLEMENTATION
    import pendulum

    class DateTimePendulum(SQLObject):
        col1 = DateTimeCol()

    def test_PendulumDateTime():
        setupClass(DateTimePendulum)
        _now = pendulum.now()
        dtp = DateTimePendulum(col1=_now)

        assert isinstance(dtp.col1, datetime)
        assert dtp.col1.year == _now.year
        assert dtp.col1.month == _now.month
        assert dtp.col1.day == _now.day
        assert dtp.col1.hour == _now.hour
        assert dtp.col1.minute == _now.minute
        assert int(dtp.col1.second) == int(_now.second)


if zope_datetime_available:
    col.default_datetime_implementation = ZOPE_DATETIME_IMPLEMENTATION
    from DateTime import DateTime as zopeDateTime

    class DateTime3(SQLObject):
        col1 = DateTimeCol()

    def test_ZopeDateTime():
        setupClass(DateTime3)
        _now = zopeDateTime()
        dt3 = DateTime3(col1=_now)

        assert isinstance(dt3.col1, col.zopeDateTimeType)
        assert dt3.col1.year() == _now.year()
        assert dt3.col1.month() == _now.month()
        assert dt3.col1.day() == _now.day()
        assert dt3.col1.hour() == _now.hour()
        assert dt3.col1.minute() == _now.minute()
        assert int(dt3.col1.second()) == int(_now.second())



import pytest
from sqlobject import RelatedJoin, SQLObject, SQLRelatedJoin, StringCol, \
    ForeignKey
from sqlobject.sqlbuilder import Alias
from sqlobject.tests.dbtest import setupClass, supports


class Fighter(SQLObject):
    class sqlmeta:
        idName = 'fighter_id'  # test on a non-standard way
    name = StringCol()
    tourtments = RelatedJoin('Tourtment')


class Tourtment(SQLObject):
    class sqlmeta:
        table = 'competition'  # test on a non-standard way
    name = StringCol()
    fightersAsList = RelatedJoin('Fighter')
    fightersAsSResult = SQLRelatedJoin('Fighter')


def createAllTables():
    setupClass(Fighter)
    setupClass(Tourtment)


def createData():
    createAllTables()
    # create some tourtments
    t1 = Tourtment(name='Tourtment #1')
    t2 = Tourtment(name='Tourtment #2')
    t3 = Tourtment(name='Tourtment #3')
    # create some fighters
    gokou = Fighter(name='gokou')
    vegeta = Fighter(name='vegeta')
    gohan = Fighter(name='gohan')
    trunks = Fighter(name='trunks')
    # relating them
    t1.addFighter(gokou)
    t1.addFighter(vegeta)
    t1.addFighter(gohan)
    t2.addFighter(gokou)
    t2.addFighter(vegeta)
    t2.addFighter(trunks)
    t3.addFighter(gohan)
    t3.addFighter(trunks)
    return t1, t2, t3, gokou, vegeta, gohan, trunks


def test_1():
    t1, t2, t3, gokou, vegeta, gohan, trunks = createData()
    for i, j in zip(t1.fightersAsList, t1.fightersAsSResult):
        assert i is j
    assert len(t2.fightersAsList) == t2.fightersAsSResult.count()


def test_related_join_transaction():
    if not supports('transactions'):
        pytest.skip("Transactions aren't supported")
    createAllTables()
    trans = Tourtment._connection.transaction()
    try:
        t1 = Tourtment(name='Tourtment #1', connection=trans)
        t1.addFighter(Fighter(name='Jim', connection=trans))
        assert t1.fightersAsSResult.count() == 1
        assert t1.fightersAsSResult[0]._connection == trans
    finally:
        trans.commit(True)
        Tourtment._connection.autoCommit = True


def test_related_join_filter():
    t1, t2, t3, gokou, vegeta, gohan, trunks = createData()
    filteredFighters = t1.fightersAsSResult.filter(
        Fighter.q.name.startswith('go')
    ).orderBy('id')
    for i, j in zip(filteredFighters, [gokou, gohan]):
        assert i is j


class RecursiveGroup(SQLObject):
    name = StringCol(length=255, unique=True)
    subgroups = SQLRelatedJoin(
        'RecursiveGroup',
        otherColumn='group_id',
        intermediateTable='rec_group_map',
        createRelatedTable=False,
    )


class RecGroupMap(SQLObject):
    recursive_group = ForeignKey('RecursiveGroup')
    group = ForeignKey('RecursiveGroup')


def test_rec_group():
    setupClass([RecursiveGroup, RecGroupMap])
    a = RecursiveGroup(name='a')
    a1 = RecursiveGroup(name='a1')
    a.addRecursiveGroup(a1)
    a2 = RecursiveGroup(name='a2')
    a.addRecursiveGroup(a2)

    assert sorted(a.subgroups, key=lambda x: x.name) == [a1, a2]

    pytest.raises(
        ValueError,
        a.subgroups.filter,
        RecursiveGroup.q.name == 'a1',
    )

    rgroupAlias = Alias(RecursiveGroup, '_SO_SQLRelatedJoin_OtherTable')
    assert list(a.subgroups.filter(rgroupAlias.q.name == 'a1')) == [a1]



from sqlobject import IntCol, SQLObject, StringCol, sqlmeta
from sqlobject.tests.dbtest import inserts, setupClass

# Test more complex orderBy clauses


class ComplexNames(SQLObject):

    class sqlmeta(sqlmeta):
        table = 'names_table'
        defaultOrder = ['lastName', 'firstName', 'phone', 'age']

    firstName = StringCol(length=30)
    lastName = StringCol(length=30)
    phone = StringCol(length=11)
    age = IntCol()


def setupComplexNames():
    setupClass(ComplexNames)
    inserts(ComplexNames, [('aj', 'baker', '555-444-333', 34),
                           ('joe', 'robbins', '444-555-333', 34),
                           ('tim', 'jackson', '555-444-222', 32),
                           ('joe', 'baker', '222-111-000', 24),
                           ('zoe', 'robbins', '444-555-333', 46)],
            schema='firstName lastName phone age')


def nameList(names):
    result = []
    for name in names:
        result.append('%s %s' % (name.firstName, name.lastName))
    return result


def firstList(names):
    return [n.firstName for n in names]


def test_defaultComplexOrder():
    setupComplexNames()
    assert nameList(ComplexNames.select()) == \
        ['aj baker', 'joe baker',
         'tim jackson', 'joe robbins',
         'zoe robbins']


def test_complexOrders():
    setupComplexNames()
    assert nameList(ComplexNames.select().orderBy(['age', 'phone',
                                                   'firstName',
                                                   'lastName'])) == \
        ['joe baker', 'tim jackson',
         'joe robbins', 'aj baker',
         'zoe robbins']
    assert nameList(ComplexNames.select().orderBy(['-age', 'phone',
                                                   'firstName',
                                                   'lastName'])) == \
        ['zoe robbins', 'joe robbins',
         'aj baker', 'tim jackson',
         'joe baker']
    assert nameList(ComplexNames.select().orderBy(['age', '-phone',
                                                   'firstName',
                                                   'lastName'])) == \
        ['joe baker', 'tim jackson',
         'aj baker', 'joe robbins',
         'zoe robbins']
    assert nameList(ComplexNames.select().orderBy(['-firstName', 'phone',
                                                   'lastName', 'age'])) == \
        ['zoe robbins', 'tim jackson',
         'joe baker', 'joe robbins',
         'aj baker']
    assert nameList(ComplexNames.select().orderBy(['-firstName', '-phone',
                                                   'lastName', 'age'])) == \
        ['zoe robbins', 'tim jackson',
         'joe robbins', 'joe baker',
         'aj baker']



"""
Declarative objects.

Declarative objects have a simple protocol: you can use classes in
lieu of instances and they are equivalent, and any keyword arguments
you give to the constructor will override those instance variables.
(So if a class is received, we'll simply instantiate an instance with
no arguments).

You can provide a variable __unpackargs__ (a list of strings), and if
the constructor is called with non-keyword arguments they will be
interpreted as the given keyword arguments.

If __unpackargs__ is ('*', name), then all the arguments will be put
in a variable by that name.

You can define a __classinit__(cls, new_attrs) method, which will be
called when the class is created (including subclasses).  Note: you
can't use super() in __classinit__ because the class isn't bound to a
name.  As an analog to __classinit__, Declarative adds
__instanceinit__ which is called with the same argument (new_attrs).
This is like __init__, but after __unpackargs__ and other factors have
been taken into account.

If __mutableattributes__ is defined as a sequence of strings, these
attributes will not be shared between superclasses and their
subclasses.  E.g., if you have a class variable that contains a list
and you append to that list, changes to subclasses will effect
superclasses unless you add the attribute here.

Also defines classinstancemethod, which acts as either a class method
or an instance method depending on where it is called.
"""

import copy
from . import events
from sqlobject.compat import with_metaclass

import itertools
counter = itertools.count()

__all__ = ('classinstancemethod', 'DeclarativeMeta', 'Declarative')


class classinstancemethod(object):
    """
    Acts like a class method when called from a class, like an
    instance method when called by an instance.  The method should
    take two arguments, 'self' and 'cls'; one of these will be None
    depending on how the method was called.
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, type=None):
        return _methodwrapper(self.func, obj=obj, type=type)


class _methodwrapper(object):

    def __init__(self, func, obj, type):
        self.func = func
        self.obj = obj
        self.type = type

    def __call__(self, *args, **kw):
        assert 'self' not in kw and 'cls' not in kw, (
            "You cannot use 'self' or 'cls' arguments to a "
            "classinstancemethod")
        return self.func(*((self.obj, self.type) + args), **kw)

    def __repr__(self):
        if self.obj is None:
            return ('<bound class method %s.%s>'
                    % (self.type.__name__, self.func.__name__))
        else:
            return ('<bound method %s.%s of %r>'
                    % (self.type.__name__, self.func.__name__, self.obj))


class DeclarativeMeta(type):

    def __new__(meta, class_name, bases, new_attrs):
        post_funcs = []
        early_funcs = []
        events.send(events.ClassCreateSignal,
                    bases[0], class_name, bases, new_attrs,
                    post_funcs, early_funcs)
        cls = type.__new__(meta, class_name, bases, new_attrs)
        for func in early_funcs:
            func(cls)
        if '__classinit__' in new_attrs:
            if hasattr(cls.__classinit__, '__func__'):
                cls.__classinit__ = staticmethod(cls.__classinit__.__func__)
            else:
                cls.__classinit__ = staticmethod(cls.__classinit__)
        cls.__classinit__(cls, new_attrs)
        for func in post_funcs:
            func(cls)
        return cls


class Declarative(with_metaclass(DeclarativeMeta, object)):

    __unpackargs__ = ()

    __mutableattributes__ = ()

    __restrict_attributes__ = None

    def __classinit__(cls, new_attrs):
        cls.declarative_count = next(counter)
        for name in cls.__mutableattributes__:
            if name not in new_attrs:
                setattr(cls, copy.copy(getattr(cls, name)))

    def __instanceinit__(self, new_attrs):
        if self.__restrict_attributes__ is not None:
            for name in new_attrs:
                if name not in self.__restrict_attributes__:
                    raise TypeError(
                        '%s() got an unexpected keyword argument %r'
                        % (self.__class__.__name__, name))
        for name, value in new_attrs.items():
            setattr(self, name, value)
        if 'declarative_count' not in new_attrs:
            self.declarative_count = next(counter)

    def __init__(self, *args, **kw):
        if self.__unpackargs__ and self.__unpackargs__[0] == '*':
            assert len(self.__unpackargs__) == 2, \
                "When using __unpackargs__ = ('*', varname), " \
                "you must only provide a single variable name " \
                "(you gave %r)" % self.__unpackargs__
            name = self.__unpackargs__[1]
            if name in kw:
                raise TypeError(
                    "keyword parameter '%s' was given by position and name"
                    % name)
            kw[name] = args
        else:
            if len(args) > len(self.__unpackargs__):
                raise TypeError(
                    '%s() takes at most %i arguments (%i given)'
                    % (self.__class__.__name__,
                       len(self.__unpackargs__),
                       len(args)))
            for name, arg in zip(self.__unpackargs__, args):
                if name in kw:
                    raise TypeError(
                        "keyword parameter '%s' was given by position and name"
                        % name)
                kw[name] = arg
        if '__alsocopy' in kw:
            for name, value in kw['__alsocopy'].items():
                if name not in kw:
                    if name in self.__mutableattributes__:
                        value = copy.copy(value)
                    kw[name] = value
            del kw['__alsocopy']
        self.__instanceinit__(kw)

    def __call__(self, *args, **kw):
        kw['__alsocopy'] = self.__dict__
        return self.__class__(*args, **kw)

    @classinstancemethod
    def singleton(self, cls):
        if self:
            return self
        name = '_%s__singleton' % cls.__name__
        if not hasattr(cls, name):
            setattr(cls, name, cls(declarative_count=cls.declarative_count))
        return getattr(cls, name)

    @classinstancemethod
    def __repr__(self, cls):
        if self:
            name = '%s object' % self.__class__.__name__
            v = self.__dict__.copy()
        else:
            name = '%s class' % cls.__name__
            v = cls.__dict__.copy()
        if 'declarative_count' in v:
            name = '%s %i' % (name, v['declarative_count'])
            del v['declarative_count']
        # @@: simplifying repr:
        # v = {}
        names = v.keys()
        args = []
        for n in self._repr_vars(names):
            args.append('%s=%r' % (n, v[n]))
        if not args:
            return '<%s>' % name
        else:
            return '<%s %s>' % (name, ' '.join(args))

    @staticmethod
    def _repr_vars(dictNames):
        names = [n for n in dictNames
                 if not n.startswith('_')
                 and n != 'declarative_count']
        names.sort()
        return names


def setup_attributes(cls, new_attrs):
    for name, value in new_attrs.items():
        if hasattr(value, '__addtoclass__'):
            value.__addtoclass__(cls, name)



from sqlobject import IntCol, SQLObject, StringCol
from sqlobject.compat import PY2
from sqlobject.sqlbuilder import AND, ANY, CONCAT, Delete, Insert, \
    SQLOp, Select, Union, Update, const, func, sqlrepr
from sqlobject.tests.dbtest import getConnection, raises, setupClass


class SOTestSQLBuilder(SQLObject):
    name = StringCol()
    so_value = IntCol()


def test_Select():
    setupClass(SOTestSQLBuilder)

    select1 = Select([const.id, func.MAX(const.salary)],
                     staticTables=['employees'])
    assert sqlrepr(select1) == 'SELECT id, MAX(salary) FROM employees'

    select2 = Select([SOTestSQLBuilder.q.name, SOTestSQLBuilder.q.so_value])
    assert sqlrepr(select2) == \
        'SELECT so_test_sql_builder.name, so_test_sql_builder.so_value ' \
        'FROM so_test_sql_builder'

    union = Union(select1, select2)
    assert sqlrepr(union) == \
        'SELECT id, MAX(salary) FROM employees ' \
        'UNION SELECT so_test_sql_builder.name, ' \
        'so_test_sql_builder.so_value ' \
        'FROM so_test_sql_builder'

    union = Union(SOTestSQLBuilder.select().queryForSelect())
    assert sqlrepr(union) == \
        'SELECT so_test_sql_builder.id, so_test_sql_builder.name, ' \
        'so_test_sql_builder.so_value FROM so_test_sql_builder WHERE 1 = 1'


def test_empty_AND():
    assert AND() is None
    assert AND(True) is True

    # sqlrepr() is needed because AND() returns an SQLExpression that overrides
    # comparison. The following
    #     AND('x', 'y') == "foo bar"
    # is True! (-: Eeek!
    assert sqlrepr(AND(1, 2)) == sqlrepr(SQLOp("AND", 1, 2)) == "((1) AND (2))"
    assert sqlrepr(AND(1, 2, '3'), "sqlite") == \
        sqlrepr(SQLOp("AND", 1, SQLOp("AND", 2, '3')), "sqlite") == \
        "((1) AND ((2) AND ('3')))"


def test_modulo():
    setupClass(SOTestSQLBuilder)
    assert sqlrepr(SOTestSQLBuilder.q.so_value % 2 == 0, 'mysql') == \
        "((MOD(so_test_sql_builder.so_value, 2)) = (0))"
    assert sqlrepr(SOTestSQLBuilder.q.so_value % 2 == 0, 'sqlite') == \
        "(((so_test_sql_builder.so_value) % (2)) = (0))"


def test_division():
    SOTestSQLBuilder(name='test', so_value=-11)
    assert sqlrepr(SOTestSQLBuilder.q.so_value / 4 == -2.75, 'mysql') == \
        "(((so_test_sql_builder.so_value) / (4)) = (-2.75))"
    assert sqlrepr(SOTestSQLBuilder.q.so_value // 4 == -3, 'mysql') == \
        "((FLOOR(((so_test_sql_builder.so_value) / (4)))) = (-3))"


def test_str_or_sqlrepr():
    select = Select(['id', 'name'], staticTables=['employees'],
                    where='value>0', orderBy='id')
    assert sqlrepr(select, 'sqlite') == \
        'SELECT id, name FROM employees WHERE value>0 ORDER BY id'

    select = Select(['id', 'name'], staticTables=['employees'],
                    where='value>0', orderBy='id', lazyColumns=True)
    assert sqlrepr(select, 'sqlite') == \
        'SELECT id FROM employees WHERE value>0 ORDER BY id'

    insert = Insert('employees', values={'id': 1, 'name': 'test'})
    assert sqlrepr(insert, 'sqlite') == \
        "INSERT INTO employees (id, name) VALUES (1, 'test')"

    update = Update('employees', {'name': 'test'}, where='id=1')
    assert sqlrepr(update, 'sqlite') == \
        "UPDATE employees SET name='test' WHERE id=1"

    update = Update('employees', {'name': 'test', 'age': 42}, where='id=1')
    assert sqlrepr(update, 'sqlite') == \
        "UPDATE employees SET age=42, name='test' WHERE id=1"

    delete = Delete('employees', where='id=1')
    assert sqlrepr(delete, 'sqlite') == \
        "DELETE FROM employees WHERE id=1"

    raises(TypeError, Delete, 'employees')

    delete = Delete('employees', where=None)
    assert sqlrepr(delete, 'sqlite') == \
        "DELETE FROM employees"


def test_CONCAT():
    setupClass(SOTestSQLBuilder)
    SOTestSQLBuilder(name='test', so_value=42)

    assert sqlrepr(CONCAT('a', 'b'), 'mysql') == "CONCAT('a', 'b')"
    assert sqlrepr(CONCAT('a', 'b'), 'mssql') == "'a' + 'b'"
    assert sqlrepr(CONCAT('a', 'b'), 'sqlite') == "'a' || 'b'"
    assert sqlrepr(CONCAT('prefix', SOTestSQLBuilder.q.name), 'mysql') == \
        "CONCAT('prefix', so_test_sql_builder.name)"
    assert sqlrepr(CONCAT('prefix', SOTestSQLBuilder.q.name), 'sqlite') == \
        "'prefix' || so_test_sql_builder.name"

    select = Select([CONCAT(SOTestSQLBuilder.q.name, '-suffix')],
                    staticTables=['so_test_sql_builder'])
    connection = getConnection()
    rows = connection.queryAll(connection.sqlrepr(select))
    result = rows[0][0]
    if not PY2 and not isinstance(result, str):
        result = result.decode('ascii')
    assert result == "test-suffix"


def test_ANY():
    setupClass(SOTestSQLBuilder)

    select = Select(
        [SOTestSQLBuilder.q.name],
        'value' == ANY(SOTestSQLBuilder.q.so_value),
    )

    assert sqlrepr(select, 'mysql') == \
        "SELECT so_test_sql_builder.name FROM so_test_sql_builder " \
        "WHERE (('value') = ANY (so_test_sql_builder.so_value))"



import pytest
from sqlobject import SQLObject, UnicodeCol
from sqlobject.tests.dbtest import getConnection, setupClass, supports


########################################
# Schema per connection
########################################


class SOTestSchema(SQLObject):
    foo = UnicodeCol(length=200)


def test_connection_schema():
    if not supports('schema'):
        pytest.skip("schemas aren't supported")
    conn = getConnection()
    conn.schema = None
    conn.query('CREATE SCHEMA test')
    conn.schema = 'test'
    conn.query('SET search_path TO test')
    setupClass(SOTestSchema)
    assert SOTestSchema._connection is conn
    SOTestSchema(foo='bar')
    assert conn.queryAll("SELECT * FROM test.so_test_schema")
    conn.schema = None
    conn.query('SET search_path TO public')



import pytest
from sqlobject import SQLObject, IntCol
from sqlobject.sqlbuilder import Select, ANY
from sqlobject.tests.dbtest import getConnection, setupClass


try:
    connection = getConnection()
except (AttributeError, NameError):
    # The module was imported during documentation building
    pass
else:
    if connection.dbName != "mysql":
        pytestmark = pytest.mark.skip("These tests require MySQL")


class SOTestSOListMySQL(SQLObject):
    pass


def test_list_databases():
    assert connection.db in connection.listDatabases()


def test_list_tables():
    setupClass(SOTestSOListMySQL)
    assert SOTestSOListMySQL.sqlmeta.table in connection.listTables()


class SOTestANY(SQLObject):
    value = IntCol()


def test_ANY():
    setupClass(SOTestANY)
    SOTestANY(value=10)
    SOTestANY(value=20)
    SOTestANY(value=30)
    assert len(list(SOTestANY.select(
        SOTestANY.q.value > ANY(Select([SOTestANY.q.value]))))) == 2


class SOTestMySQLidSize(SQLObject):
    class sqlmeta:
        idSize = 'BIG'


def test_idSize():
    assert 'id BIGINT PRIMARY KEY AUTO_INCREMENT' \
        in SOTestMySQLidSize.createTableSQL(connection=connection)[0]



import pytest
from sqlobject import BLOBCol, SQLObject
from sqlobject.compat import PY2
from sqlobject.tests.dbtest import setupClass, supports


########################################
# BLOB columns
########################################


class ImageData(SQLObject):
    image = BLOBCol(default=b'emptydata', length=256)


def test_BLOBCol():
    if not supports('blobData'):
        pytest.skip("blobData isn't supported")
    setupClass(ImageData)
    if PY2:
        data = ''.join([chr(x) for x in range(256)])
    else:
        data = bytes(range(256))

    prof = ImageData(image=data)
    iid = prof.id

    ImageData._connection.cache.clear()

    prof2 = ImageData.get(iid)
    assert prof2.image == data

    ImageData(image=b'string')
    assert ImageData.selectBy(image=b'string').count() == 1



from sqlobject import IntCol, SQLObject, UnicodeCol
from sqlobject.compat import PY2
from sqlobject.tests.dbtest import setupClass


########################################
# Unicode columns
########################################


class SOTestUnicode(SQLObject):
    so_count = IntCol(alternateID=True)
    col = UnicodeCol(alternateID=True, length=100)


data = [u'\u00f0', u'test', 'ascii test']
items = []


def setup():
    global items
    items = []
    setupClass(SOTestUnicode)
    for i, s in enumerate(data):
        items.append(SOTestUnicode(so_count=i, col=s))


def test_create():
    setup()
    for s, item in zip(data, items):
        assert item.col == s

    conn = SOTestUnicode._connection
    if PY2:
        rows = conn.queryAll("""
        SELECT so_count, col
        FROM so_test_unicode
        ORDER BY so_count
        """)
        for so_count, col in rows:
            if not isinstance(col, bytes):
                col = col.encode('utf-8')
            assert data[so_count].encode('utf-8') == col
    else:
        rows = conn.queryAll("""
        SELECT so_count, col
        FROM so_test_unicode
        ORDER BY so_count
        """)
        # On python 3, everthings already decoded to unicode
        for so_count, col in rows:
            assert data[so_count] == col


def test_select():
    setup()
    for i, value in enumerate(data):
        rows = list(SOTestUnicode.select(SOTestUnicode.q.col == value))
        assert len(rows) == 1
        if PY2:
            rows = list(SOTestUnicode.select(SOTestUnicode.q.col == value))
            assert len(rows) == 1
        rows = list(SOTestUnicode.selectBy(col=value))
        assert len(rows) == 1
        if PY2:
            rows = list(SOTestUnicode.selectBy(col=value))
            assert len(rows) == 1
        row = SOTestUnicode.byCol(value)
        assert row.so_count == i

    # starts/endswith/contains
    rows = list(SOTestUnicode.select(SOTestUnicode.q.col.startswith("test")))
    assert len(rows) == 1
    rows = list(SOTestUnicode.select(SOTestUnicode.q.col.endswith("test")))
    assert len(rows) == 2
    rows = list(SOTestUnicode.select(SOTestUnicode.q.col.contains("test")))
    assert len(rows) == 2
    rows = list(SOTestUnicode.select(
        SOTestUnicode.q.col.startswith(u"\u00f0")))
    assert len(rows) == 1
    rows = list(SOTestUnicode.select(SOTestUnicode.q.col.endswith(u"\u00f0")))
    assert len(rows) == 1
    rows = list(SOTestUnicode.select(SOTestUnicode.q.col.contains(u"\u00f0")))
    assert len(rows) == 1



from sqlobject import SQLObject, StringCol
from sqlobject.tests.dbtest import setupClass


########################################
# Test hashing a column instance
########################################


class ClassHashTest(SQLObject):
    name = StringCol(length=50, alternateID=True, dbName='name_col')


def test_class_hash():
    setupClass(ClassHashTest)
    ClassHashTest(name='bob')

    b = ClassHashTest.byName('bob')
    hashed = hash(b)
    b.expire()
    b = ClassHashTest.byName('bob')
    assert hash(b) == hashed



from sqlobject import ForeignKey, IntCol, SQLObject
from sqlobject.inheritance import InheritableSQLObject
from sqlobject.tests.dbtest import setupClass


class SOTestCascade1(InheritableSQLObject):
    dummy = IntCol()


class SOTestCascade2(SOTestCascade1):
    c = ForeignKey('SOTestCascade3', cascade='null')


class SOTestCascade3(SQLObject):
    dummy = IntCol()


def test_destroySelf():
    setupClass([SOTestCascade1, SOTestCascade3, SOTestCascade2])

    c = SOTestCascade3(dummy=1)
    SOTestCascade2(cID=c.id, dummy=1)
    c.destroySelf()
